#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, re, sys
from waflib import Utils, Errors, Logs
import waflib.Node
if sys.hexversion > 0x3040000:
    import types

    class imp(object):
        new_module = lambda x: types.ModuleType(x)
else:
    import imp

HEXVERSION = 0x2001400
WAFVERSION = "2.0.20"
WAFREVISION = "668769470956da8c5b60817cb8884cd7d0f87cd4"
WAFNAME = "waf"
ABI = 20
DBFILE = '.wafpickle-%s-%d-%d' % (sys.platform, sys.hexversion, ABI)
APPNAME = 'APPNAME'
VERSION = 'VERSION'
TOP = 'top'
OUT = 'out'
WSCRIPT_FILE = 'wscript'
launch_dir = ''
run_dir = ''
top_dir = ''
out_dir = ''
waf_dir = ''
default_encoding = Utils.console_encoding()
g_module = None
STDOUT = 1
STDERR = -1
BOTH = 0
classes = []


def create_context(cmd_name, *k, **kw):
    for x in classes:
        if x.cmd == cmd_name:
            return x(*k, **kw)
    ctx = Context(*k, **kw)
    ctx.fun = cmd_name
    return ctx


class store_context(type):
    def __init__(cls, name, bases, dct):
        super(store_context, cls).__init__(name, bases, dct)
        name = cls.__name__
        if name in ('ctx', 'Context'):
            return
        try:
            cls.cmd
        except AttributeError:
            raise Errors.WafError('Missing command for the context class %r (cmd)' % name)
        if not getattr(cls, 'fun', None):
            cls.fun = cls.cmd
        classes.insert(0, cls)


ctx = store_context('ctx', (object,), {})


class Context(ctx):
    errors = Errors
    tools = {}

    def __init__(self, **kw):
        try:
            rd = kw['run_dir']
        except KeyError:
            rd = run_dir
        self.node_class = type('Nod3', (waflib.Node.Node,), {})
        self.node_class.__module__ = 'waflib.Node'
        self.node_class.ctx = self
        self.root = self.node_class('', None)
        self.cur_script = None
        self.path = self.root.find_dir(rd)
        self.stack_path = []
        self.exec_dict = {'ctx': self, 'conf': self, 'bld': self, 'opt': self}
        self.logger = None

    def finalize(self):
        try:
            logger = self.logger
        except AttributeError:
            pass
        else:
            Logs.free_logger(logger)
            delattr(self, 'logger')

    def load(self, tool_list, *k, **kw):
        tools = Utils.to_list(tool_list)
        path = Utils.to_list(kw.get('tooldir', ''))
        with_sys_path = kw.get('with_sys_path', True)
        for t in tools:
            module = load_tool(t, path, with_sys_path=with_sys_path)
            fun = getattr(module, kw.get('name', self.fun), None)
            if fun:
                fun(self)

    def execute(self):
        self.recurse([os.path.dirname(g_module.root_path)])

    def pre_recurse(self, node):
        self.stack_path.append(self.cur_script)
        self.cur_script = node
        self.path = node.parent

    def post_recurse(self, node):
        self.cur_script = self.stack_path.pop()
        if self.cur_script:
            self.path = self.cur_script.parent

    def recurse(self, dirs, name=None, mandatory=True, once=True, encoding=None):
        try:
            cache = self.recurse_cache
        except AttributeError:
            cache = self.recurse_cache = {}
        for d in Utils.to_list(dirs):
            if not os.path.isabs(d):
                d = os.path.join(self.path.abspath(), d)
            WSCRIPT = os.path.join(d, WSCRIPT_FILE)
            WSCRIPT_FUN = WSCRIPT + '_' + (name or self.fun)
            node = self.root.find_node(WSCRIPT_FUN)
            if node and (not once or node not in cache):
                cache[node] = True
                self.pre_recurse(node)
                try:
                    function_code = node.read('r', encoding)
                    exec(compile(function_code, node.abspath(), 'exec'), self.exec_dict)
                finally:
                    self.post_recurse(node)
            elif not node:
                node = self.root.find_node(WSCRIPT)
                tup = (node, name or self.fun)
                if node and (not once or tup not in cache):
                    cache[tup] = True
                    self.pre_recurse(node)
                    try:
                        wscript_module = load_module(node.abspath(), encoding=encoding)
                        user_function = getattr(wscript_module, (name or self.fun), None)
                        if not user_function:
                            if not mandatory:
                                continue
                            raise Errors.WafError('No function %r defined in %s' % (name or self.fun, node.abspath()))
                        user_function(self)
                    finally:
                        self.post_recurse(node)
                elif not node:
                    if not mandatory:
                        continue
                    try:
                        os.listdir(d)
                    except OSError:
                        raise Errors.WafError('Cannot read the folder %r' % d)
                    raise Errors.WafError('No wscript file in directory %s' % d)

    def log_command(self, cmd, kw):
        if Logs.verbose:
            fmt = os.environ.get('WAF_CMD_FORMAT')
            if fmt == 'string':
                if not isinstance(cmd, str):
                    cmd = Utils.shell_escape(cmd)
            Logs.debug('runner: %r', cmd)
            Logs.debug('runner_env: kw=%s', kw)

    def exec_command(self, cmd, **kw):
        subprocess = Utils.subprocess
        kw['shell'] = isinstance(cmd, str)
        self.log_command(cmd, kw)
        if self.logger:
            self.logger.info(cmd)
        if 'stdout' not in kw:
            kw['stdout'] = subprocess.PIPE
        if 'stderr' not in kw:
            kw['stderr'] = subprocess.PIPE
        if Logs.verbose and not kw['shell'] and not Utils.check_exe(cmd[0]):
            raise Errors.WafError('Program %s not found!' % cmd[0])
        cargs = {}
        if 'timeout' in kw:
            if sys.hexversion >= 0x3030000:
                cargs['timeout'] = kw['timeout']
                if not 'start_new_session' in kw:
                    kw['start_new_session'] = True
            del kw['timeout']
        if 'input' in kw:
            if kw['input']:
                cargs['input'] = kw['input']
                kw['stdin'] = subprocess.PIPE
            del kw['input']
        if 'cwd' in kw:
            if not isinstance(kw['cwd'], str):
                kw['cwd'] = kw['cwd'].abspath()
        encoding = kw.pop('decode_as', default_encoding)
        try:
            ret, out, err = Utils.run_process(cmd, kw, cargs)
        except Exception as e:
            raise Errors.WafError('Execution failure: %s' % str(e), ex=e)
        if out:
            if not isinstance(out, str):
                out = out.decode(encoding, errors='replace')
            if self.logger:
                self.logger.debug('out: %s', out)
            else:
                Logs.info(out, extra={'stream': sys.stdout, 'c1': ''})
        if err:
            if not isinstance(err, str):
                err = err.decode(encoding, errors='replace')
            if self.logger:
                self.logger.error('err: %s' % err)
            else:
                Logs.info(err, extra={'stream': sys.stderr, 'c1': ''})
        return ret

    def cmd_and_log(self, cmd, **kw):
        subprocess = Utils.subprocess
        kw['shell'] = isinstance(cmd, str)
        self.log_command(cmd, kw)
        quiet = kw.pop('quiet', None)
        to_ret = kw.pop('output', STDOUT)
        if Logs.verbose and not kw['shell'] and not Utils.check_exe(cmd[0]):
            raise Errors.WafError('Program %r not found!' % cmd[0])
        kw['stdout'] = kw['stderr'] = subprocess.PIPE
        if quiet is None:
            self.to_log(cmd)
        cargs = {}
        if 'timeout' in kw:
            if sys.hexversion >= 0x3030000:
                cargs['timeout'] = kw['timeout']
                if not 'start_new_session' in kw:
                    kw['start_new_session'] = True
            del kw['timeout']
        if 'input' in kw:
            if kw['input']:
                cargs['input'] = kw['input']
                kw['stdin'] = subprocess.PIPE
            del kw['input']
        if 'cwd' in kw:
            if not isinstance(kw['cwd'], str):
                kw['cwd'] = kw['cwd'].abspath()
        encoding = kw.pop('decode_as', default_encoding)
        try:
            ret, out, err = Utils.run_process(cmd, kw, cargs)
        except Exception as e:
            raise Errors.WafError('Execution failure: %s' % str(e), ex=e)
        if not isinstance(out, str):
            out = out.decode(encoding, errors='replace')
        if not isinstance(err, str):
            err = err.decode(encoding, errors='replace')
        if out and quiet != STDOUT and quiet != BOTH:
            self.to_log('out: %s' % out)
        if err and quiet != STDERR and quiet != BOTH:
            self.to_log('err: %s' % err)
        if ret:
            e = Errors.WafError('Command %r returned %r' % (cmd, ret))
            e.returncode = ret
            e.stderr = err
            e.stdout = out
            raise e
        if to_ret == BOTH:
            return (out, err)
        elif to_ret == STDERR:
            return err
        return out

    def fatal(self, msg, ex=None):
        if self.logger:
            self.logger.info('from %s: %s' % (self.path.abspath(), msg))
        try:
            logfile = self.logger.handlers[0].baseFilename
        except AttributeError:
            pass
        else:
            if os.environ.get('WAF_PRINT_FAILURE_LOG'):
                msg = 'Log from (%s):\n%s\n' % (logfile, Utils.readf(logfile))
            else:
                msg = '%s\n(complete log in %s)' % (msg, logfile)
        raise self.errors.ConfigurationError(msg, ex=ex)

    def to_log(self, msg):
        if not msg:
            return
        if self.logger:
            self.logger.info(msg)
        else:
            sys.stderr.write(str(msg))
            sys.stderr.flush()

    def msg(self, *k, **kw):
        try:
            msg = kw['msg']
        except KeyError:
            msg = k[0]
        self.start_msg(msg, **kw)
        try:
            result = kw['result']
        except KeyError:
            result = k[1]
        color = kw.get('color')
        if not isinstance(color, str):
            color = result and 'GREEN' or 'YELLOW'
        self.end_msg(result, color, **kw)

    def start_msg(self, *k, **kw):
        if kw.get('quiet'):
            return
        msg = kw.get('msg') or k[0]
        try:
            if self.in_msg:
                self.in_msg += 1
                return
        except AttributeError:
            self.in_msg = 0
        self.in_msg += 1
        try:
            self.line_just = max(self.line_just, len(msg))
        except AttributeError:
            self.line_just = max(40, len(msg))
        for x in (self.line_just * '-', msg):
            self.to_log(x)
        Logs.pprint('NORMAL', "%s :" % msg.ljust(self.line_just), sep='')

    def end_msg(self, *k, **kw):
        if kw.get('quiet'):
            return
        self.in_msg -= 1
        if self.in_msg:
            return
        result = kw.get('result') or k[0]
        defcolor = 'GREEN'
        if result is True:
            msg = 'ok'
        elif not result:
            msg = 'not found'
            defcolor = 'YELLOW'
        else:
            msg = str(result)
        self.to_log(msg)
        try:
            color = kw['color']
        except KeyError:
            if len(k) > 1 and k[1] in Logs.colors_lst:
                color = k[1]
            else:
                color = defcolor
        Logs.pprint(color, msg)

    def load_special_tools(self, var, ban=[]):
        if os.path.isdir(waf_dir):
            lst = self.root.find_node(waf_dir).find_node('waflib/extras').ant_glob(var)
            for x in lst:
                if not x.name in ban:
                    load_tool(x.name.replace('.py', ''))
        else:
            from zipfile import PyZipFile
            waflibs = PyZipFile(waf_dir)
            lst = waflibs.namelist()
            for x in lst:
                if not re.match('waflib/extras/%s' % var.replace('*', '.*'), var):
                    continue
                f = os.path.basename(x)
                doban = False
                for b in ban:
                    r = b.replace('*', '.*')
                    if re.match(r, f):
                        doban = True
                if not doban:
                    f = f.replace('.py', '')
                    load_tool(f)


cache_modules = {}


def load_module(path, encoding=None):
    try:
        return cache_modules[path]
    except KeyError:
        pass
    module = imp.new_module(WSCRIPT_FILE)
    try:
        code = Utils.readf(path, m='r', encoding=encoding)
    except EnvironmentError:
        raise Errors.WafError('Could not read the file %r' % path)
    module_dir = os.path.dirname(path)
    sys.path.insert(0, module_dir)
    try:
        exec(compile(code, path, 'exec'), module.__dict__)
    finally:
        sys.path.remove(module_dir)
    cache_modules[path] = module
    return module


def load_tool(tool, tooldir=None, ctx=None, with_sys_path=True):
    if tool == 'java':
        tool = 'javaw'
    else:
        tool = tool.replace('++', 'xx')
    if not with_sys_path:
        back_path = sys.path
        sys.path = []
    try:
        if tooldir:
            assert isinstance(tooldir, list)
            sys.path = tooldir + sys.path
            try:
                __import__(tool)
            except ImportError as e:
                e.waf_sys_path = list(sys.path)
                raise
            finally:
                for d in tooldir:
                    sys.path.remove(d)
            ret = sys.modules[tool]
            Context.tools[tool] = ret
            return ret
        else:
            if not with_sys_path:
                sys.path.insert(0, waf_dir)
            try:
                for x in ('waflib.Tools.%s', 'waflib.extras.%s', 'waflib.%s', '%s'):
                    try:
                        __import__(x % tool)
                        break
                    except ImportError:
                        x = None
                else:
                    __import__(tool)
            except ImportError as e:
                e.waf_sys_path = list(sys.path)
                raise
            finally:
                if not with_sys_path:
                    sys.path.remove(waf_dir)
            ret = sys.modules[x % tool]
            Context.tools[tool] = ret
            return ret
    finally:
        if not with_sys_path:
            sys.path += back_path



available_optional_dependencies = []
try:
    import email  # noqa: F401
    available_optional_dependencies.append("email")
except ImportError:
    pass
try:
    import gzip  # noqa: F401
    available_optional_dependencies.append("gzip")
except ImportError:
    pass
try:
    import pstats  # noqa: F401
    available_optional_dependencies.append("pstats")
except ImportError:
    pass
print("Available dependencies:", *available_optional_dependencies)



#-----------------------------------------------------------------------------
# Copyright (c) 2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import sys
import subprocess

import pytest

from PyInstaller.compat import is_win


# Test that bootloader is able to fully resolve a symlinked executable, on all OSes.
def test_bootloader_symlinked_executable(pyi_builder, tmp_path):
    pyi_builder.test_source(
        """
        import sys

        print(f"sys.argv={sys.argv}")
        print(f"sys.executable={sys.executable}")
        """
    )

    # Find executable and run additional tests with symbolic links
    print('------- Looking for executable. -------', file=sys.stderr)
    exes = pyi_builder._find_executables('test_source')
    assert len(exes) == 1
    executable = exes[0]
    print(f"True executable location: {executable}")

    exe_suffix = '.exe' if is_win else ''

    # Case #1: a relative symbolic link in a subdirectory
    print('------- Test case #1: relative symlink in subdirectory -------', file=sys.stderr)
    subdir = tmp_path / 'subdir'
    subdir.mkdir()

    symlink1 = subdir / f'relative-symlink-in-subdir{exe_suffix}'
    rel_link = os.path.relpath(executable, subdir)  # Compute relative link
    print(f"Relative symbolic link: {symlink1} -> {rel_link}", file=sys.stderr)

    _create_symlink(rel_link, symlink1)

    subprocess.check_call([symlink1])

    # Case #2: an absolute symbolic link pointing to relative symbolic link from case #1.
    print('------- Test case #2: absolute symlink pointing to relative symlink -------', file=sys.stderr)
    symlink2 = tmp_path / f'absolute-symlink{exe_suffix}'
    print(f"Absolute symbolic link: {symlink2} -> {symlink1}", file=sys.stderr)

    _create_symlink(symlink1, symlink2)

    subprocess.check_call([symlink2])


# Wrapper for os.symlink that skips the test if symlink cannot be created on Windows.
def _create_symlink(src, dst, *args, **kwargs):
    # On Windows, symlinks need to use Windows-style backslashes, even if we are under msys2/mingw python that uses
    # POSIX-style separators. `src` might be a Path-like object, so explicitly convert it to string.
    if is_win and os.sep == '/':
        src = str(src).replace(os.sep, '\\')

    try:
        os.symlink(src, dst, **kwargs)
    except OSError:
        if is_win:
            pytest.skip("OS does not support creation of symbolic links.")
        else:
            raise


def _create_data(tmp_path, orig_filename, link_filename):
    # Create data directory
    data_path = tmp_path / "data"

    # Create original file
    abs_orig_filename = data_path / orig_filename
    abs_orig_filename.parent.mkdir(parents=True, exist_ok=True)
    abs_orig_filename.write_text("secret", encoding='utf-8')

    # Create symbolic link
    abs_linked_filename = data_path / link_filename
    abs_linked_filename.parent.mkdir(parents=True, exist_ok=True)
    rel_orig_filename = os.path.relpath(abs_orig_filename, abs_linked_filename.parent)

    _create_symlink(rel_orig_filename, abs_linked_filename)


# Collect both symbolic link and file
def test_symlinks__samedir__symlink_and_file(tmp_path, script_dir, pyi_builder):
    # Create test data
    _create_data(tmp_path, 'orig_file.txt', 'linked_file.txt')

    # Collect both original file and symbolic link
    data_dir = tmp_path / 'data'
    add_data_args = [
        '--add-data', f"{data_dir/'orig_file.txt'}:data",
        '--add-data', f"{data_dir/'linked_file.txt'}:data",
    ]  # yapf: disable

    # Run test script in 'samedir' mode
    pyi_builder.test_script(
        script_dir / 'pyi_symlinks_test.py',
        pyi_args=add_data_args,
        app_args=['samedir'],
    )


# Same as test_symlinks__samedir__symlink_and_file, except that we collect whole data directory instead of
# individual files.
def test_symlinks__samedir__wholedir(tmp_path, script_dir, pyi_builder):
    # Create test data
    _create_data(tmp_path, 'orig_file.txt', 'linked_file.txt')

    # Collect whole data directory
    data_dir = tmp_path / 'data'
    add_data_args = [
        '--add-data', f"{data_dir}:data",
    ]  # yapf: disable

    # Run test script in 'samedir' mode
    pyi_builder.test_script(
        script_dir / 'pyi_symlinks_test.py',
        pyi_args=add_data_args,
        app_args=['samedir'],
    )


# Collect only symbolic link
def test_symlinks__samedir__symlink_only(tmp_path, script_dir, pyi_builder):
    # Create test data
    _create_data(tmp_path, 'orig_file.txt', 'linked_file.txt')

    # Collect only symbolic link
    data_dir = tmp_path / 'data'
    add_data_args = [
        '--add-data', f"{data_dir / 'linked_file.txt'}:data",
    ]  # yapf: disable

    # Run test script in 'samedir/link-only' mode
    pyi_builder.test_script(
        script_dir / 'pyi_symlinks_test.py',
        pyi_args=add_data_args,
        app_args=['samedir', '--link-only'],
    )


# Collect symbolic link and original file, but place the latter somewhere else
def test_symlinks__samedir__orig_elsewhere(tmp_path, script_dir, pyi_builder):
    # Create test data
    _create_data(tmp_path, 'orig_file.txt', 'linked_file.txt')

    # Collect only symbolic link
    data_dir = tmp_path / 'data'
    add_data_args = [
        '--add-data', f"{data_dir / 'orig_file.txt'}:data2",
        '--add-data', f"{data_dir / 'linked_file.txt'}:data",
    ]  # yapf: disable

    # Run test script in 'samedir/link-only' mode (this should effectively be the same as 'link-only' scenario)
    pyi_builder.test_script(
        script_dir / 'pyi_symlinks_test.py',
        pyi_args=add_data_args,
        app_args=['samedir', '--link-only'],
    )


# Collect symbolic link and original file; collect original multiple times, in original and alternative locations.
def test_symlinks__samedir__orig_multiple(tmp_path, script_dir, pyi_builder):
    # Create test data
    _create_data(tmp_path, 'orig_file.txt', 'linked_file.txt')

    # Collect only symbolic link
    data_dir = tmp_path / 'data'
    add_data_args = [
        '--add-data', f"{data_dir / 'orig_file.txt'}:data3",
        '--add-data', f"{data_dir / 'orig_file.txt'}:data",
        '--add-data', f"{data_dir / 'orig_file.txt'}:data2",
        '--add-data', f"{data_dir / 'linked_file.txt'}:data",
    ]  # yapf: disable

    # Run test script in 'samedir' mode
    pyi_builder.test_script(
        script_dir / 'pyi_symlinks_test.py',
        pyi_args=add_data_args,
        app_args=['samedir'],
    )


# Test symbolic link pointing to a subdirectory
def test_symlinks__subdir__wholedir(tmp_path, script_dir, pyi_builder):
    # Create test data
    _create_data(tmp_path, os.path.join('subdir', 'orig_file.txt'), 'linked_file.txt')

    # Collect whole data directory
    data_dir = tmp_path / 'data'
    add_data_args = [
        '--add-data', f"{data_dir}:data",
    ]  # yapf: disable

    # Run test script in 'subdir' mode
    pyi_builder.test_script(
        script_dir / 'pyi_symlinks_test.py',
        pyi_args=add_data_args,
        app_args=['subdir'],
    )


# Test symbolic link pointing to a subdirectory - symlink only
def test_symlinks__subdir__symlink_only(tmp_path, script_dir, pyi_builder):
    # Create test data
    _create_data(tmp_path, os.path.join('subdir', 'orig_file.txt'), 'linked_file.txt')

    # Collect only symbolic link
    data_dir = tmp_path / 'data'
    add_data_args = [
        '--add-data', f"{data_dir / 'linked_file.txt'}:data",
    ]  # yapf: disable

    # Run test script in 'subdir/link-only' mode
    pyi_builder.test_script(
        script_dir / 'pyi_symlinks_test.py',
        pyi_args=add_data_args,
        app_args=['subdir', '--link-only'],
    )


# Test symbolic link pointing to a parent directory
def test_symlinks__parentdir__wholedir(tmp_path, script_dir, pyi_builder):
    # Create test data
    _create_data(tmp_path, 'orig_file.txt', os.path.join('subdir', 'linked_file.txt'))

    # Collect whole data directory
    data_dir = tmp_path / 'data'
    add_data_args = [
        '--add-data', f"{data_dir}:data",
    ]  # yapf: disable

    # Run test script in 'parentdir' mode
    pyi_builder.test_script(
        script_dir / 'pyi_symlinks_test.py',
        pyi_args=add_data_args,
        app_args=['parentdir'],
    )


# Test symbolic link pointing to a parent directory - symlink only
def test_symlinks__parentdir__symlink_only(tmp_path, script_dir, pyi_builder):
    # Create test data
    _create_data(tmp_path, 'orig_file.txt', os.path.join('subdir', 'linked_file.txt'))

    # Collect only symbolic link
    data_dir = tmp_path / 'data'
    add_data_args = [
        '--add-data', f"{data_dir / 'subdir' / 'linked_file.txt'}:{os.path.join('data', 'subdir')}",
    ]  # yapf: disable

    # Run test script in 'parentdir/link-only' mode
    pyi_builder.test_script(
        script_dir / 'pyi_symlinks_test.py',
        pyi_args=add_data_args,
        app_args=['parentdir', '--link-only'],
    )


# The tests below verify the behavior when collecting chained symbolic links: file_d -> file_c -> file_b -> file_a.
#
# This reproduces the scenario found under Homebrew python environment on macOS, where shared libraries for `wxwidgets`
# have the following layout:
#  * libwx_baseu-3.2.0.2.1.dylib
#  * libwx_baseu-3.2.0.dylib -> libwx_baseu-3.2.0.2.1.dylib
#  * libwx_baseu-3.2.dylib -> libwx_baseu-3.2.0.dylib
#
# And only `libwx_baseu-3.2.0.2.1.dylib` and `libwx_baseu-3.2.dylib` end up referenced by other binaries, so the
# "intermediate" link is not collected. So to prevent the symlink to be collected as a duplicated copy (which leads to
# problems on macOS...), PyInstaller needs to be able to rewrite the link to "jump over" the missing intermediate link.


# Prepare the file and links for the test
def _prepare_chained_links_example(tmp_path):
    # Create data directory
    data_path = tmp_path / "data"
    data_path.mkdir()

    # Create original file: file_a
    (data_path / "file_a").write_text("secret", encoding='utf-8')

    # Create symbolic link: file_b -> file_a
    _create_symlink("file_a", data_path / "file_b")

    # Create symbolic link: file_c -> file_b
    _create_symlink("file_b", data_path / "file_c")

    # Create symbolic link: file_d -> file_c
    _create_symlink("file_c", data_path / "file_d")

    return data_path


# Helper to generate --add-data arguments for PyInstaller.
def _collect_data(src_dir, filenames, dest_dir='.'):
    for filename in filenames:
        yield '--add-data'
        yield f"{src_dir / filename}:{dest_dir}"


def test_symlinks__chained_links_abc(tmp_path, pyi_builder):
    data_path = _prepare_chained_links_example(tmp_path)
    collected_files = ['file_a', 'file_b', 'file_c']
    pyi_builder.test_source(
        """
        import sys
        import os

        # file_a should be a regular file
        file_a = os.path.join(sys._MEIPASS, "file_a")
        assert os.path.isfile(file_a), "file_a does not exist!"
        assert not os.path.islink(file_a), "file_a is a symbolic link, but should not be!"

        # file_b should be a symlink pointing to file_a
        file_b = os.path.join(sys._MEIPASS, "file_b")
        assert os.path.isfile(file_b), "file_b does not exist!"
        assert os.path.islink(file_b), "file_b is not a symbolic link!"
        assert os.readlink(file_b) == "file_a", "file_b does not point to file_a!"

        # file_c should be a symlink pointing to file_b
        file_c = os.path.join(sys._MEIPASS, "file_c")
        assert os.path.isfile(file_c), "file_c does not exist!"
        assert os.path.islink(file_c), "file_c is not a symbolic link!"
        assert os.readlink(file_c) == "file_b", "file_c does not point to file_b!"
        """,
        pyi_args=list(_collect_data(data_path, collected_files))
    )


def test_symlinks__chained_links_ab(tmp_path, pyi_builder):
    data_path = _prepare_chained_links_example(tmp_path)
    collected_files = ['file_a', 'file_b']
    pyi_builder.test_source(
        """
        import sys
        import os

        # file_a should be a regular file
        file_a = os.path.join(sys._MEIPASS, "file_a")
        assert os.path.isfile(file_a), "file_a does not exist!"
        assert not os.path.islink(file_a), "file_a is a symbolic link, but should not be!"

        # file_b should be a symlink pointing to file_a
        file_b = os.path.join(sys._MEIPASS, "file_b")
        assert os.path.isfile(file_b), "file_b does not exist!"
        assert os.path.islink(file_b), "file_b is not a symbolic link!"
        assert os.readlink(file_b) == "file_a", "file_b does not point to file_a!"
        """,
        pyi_args=list(_collect_data(data_path, collected_files))
    )


def test_symlinks__chained_links_bc(tmp_path, pyi_builder):
    data_path = _prepare_chained_links_example(tmp_path)
    collected_files = ['file_b', 'file_c']
    pyi_builder.test_source(
        """
        import sys
        import os

        # file_b should be a regular file
        file_b = os.path.join(sys._MEIPASS, "file_b")
        assert os.path.isfile(file_b), "file_b does not exist!"
        assert not os.path.islink(file_b), "file_b is a symbolic link, but should not be!"

        # file_c should be a symlink pointing to file_b
        file_c = os.path.join(sys._MEIPASS, "file_c")
        assert os.path.isfile(file_c), "file_c does not exist!"
        assert os.path.islink(file_c), "file_c is not a symbolic link!"
        assert os.readlink(file_c) == "file_b", "file_c does not point to file_b!"
        """,
        pyi_args=list(_collect_data(data_path, collected_files))
    )


# This is a special case, because PyInstaller needs to relink file_c to file_a due to missing link (file_b).
def test_symlinks__chained_links_ac(tmp_path, pyi_builder):
    data_path = _prepare_chained_links_example(tmp_path)
    collected_files = ['file_a', 'file_c']
    pyi_builder.test_source(
        """
        import sys
        import os

        # file_a should be a regular file
        file_a = os.path.join(sys._MEIPASS, "file_a")
        assert os.path.isfile(file_a), "file_a does not exist!"
        assert not os.path.islink(file_a), "file_a is a symbolic link, but should not be!"

        # file_c should be a symlink pointing to file_a
        file_c = os.path.join(sys._MEIPASS, "file_c")
        assert os.path.isfile(file_c), "file_c does not exist!"
        assert os.path.islink(file_c), "file_c is not a symbolic link!"
        assert os.readlink(file_c) == "file_a", "file_c does not point to file_a!"
        """,
        pyi_args=list(_collect_data(data_path, collected_files))
    )


# This is a special case, because PyInstaller needs to relink file_d to file_a due to missing links (file_b, file_c).
def test_symlinks__chained_links_ad(tmp_path, pyi_builder):
    data_path = _prepare_chained_links_example(tmp_path)
    collected_files = ['file_a', 'file_d']
    pyi_builder.test_source(
        """
        import sys
        import os

        # file_a should be a regular file
        file_a = os.path.join(sys._MEIPASS, "file_a")
        assert os.path.isfile(file_a), "file_a does not exist!"
        assert not os.path.islink(file_a), "file_a is a symbolic link, but should not be!"

        # file_d should be a symlink pointing to file_a
        file_d = os.path.join(sys._MEIPASS, "file_d")
        assert os.path.isfile(file_d), "file_d does not exist!"
        assert os.path.islink(file_d), "file_d is not a symbolic link!"
        assert os.readlink(file_d) == "file_a", "file_d does not point to file_a!"
        """,
        pyi_args=list(_collect_data(data_path, collected_files))
    )


# Similar to the BC test case (file_b needs to be collected as a copy), but with a missing intermediate link (file_c).
def test_symlinks__chained_links_bd(tmp_path, pyi_builder):
    data_path = _prepare_chained_links_example(tmp_path)
    collected_files = ['file_b', 'file_d']
    pyi_builder.test_source(
        """
        import sys
        import os

        # file_b should be a regular file
        file_b = os.path.join(sys._MEIPASS, "file_b")
        assert os.path.isfile(file_b), "file_b does not exist!"
        assert not os.path.islink(file_b), "file_b is a symbolic link, but should not be!"

        # file_d should be a symlink pointing to file_b
        file_d = os.path.join(sys._MEIPASS, "file_d")
        assert os.path.isfile(file_d), "file_d does not exist!"
        assert os.path.islink(file_d), "file_d is not a symbolic link!"
        assert os.readlink(file_d) == "file_b", "file_d does not point to file_b!"
        """,
        pyi_args=list(_collect_data(data_path, collected_files))
    )


# The tests below reproduce another quirk of the Homebrew python environments: the directories containing the shared
# libraries (and the corresponding versioned symbolic links) can themselves be linked to different locations, which can
# all be referenced from other binaries.
#
# For example, we have the following layout:
# * /usr/local/Cellar/wxwidgets/3.2.2.1_1/lib/libwx_baseu-3.2.0.2.1.dylib
# * /usr/local/Cellar/wxwidgets/3.2.2.1_1/lib/libwx_baseu-3.2.0.dylib -> libwx_baseu-3.2.0.2.1.dylib
# * /usr/local/Cellar/wxwidgets/3.2.2.1_1/lib/libwx_baseu-3.2.dylib -> libwx_baseu-3.2.0.dylib
# and
# * /usr/local/opt/wxwidgets/lib/libwx_baseu-3.2.0.2.1.dylib
# * /usr/local/opt/wxwidgets/lib/libwx_baseu-3.2.0.dylib -> libwx_baseu-3.2.0.2.1.dylib
# * /usr/local/opt/wxwidgets/lib/libwx_baseu-3.2.dylib -> libwx_baseu-3.2.0.dylib
# which are actually the same, because
# * /usr/local/opt/wxwidgets -> ../Cellar/wxwidgets/3.2.2.1_1
#
# Other binaries end up referencing `/usr/local/opt/wxwidgets/lib/libwx_baseu-3.2.dylib` and
# `/usr/local/Cellar/wxwidgets/3.2.2.1_1/lib/libwx_baseu-3.2.0.2.1.dylib`. So in addition the the problem with chained
# links and missing intermediate link, we also need to be able to handle the situation where the files appear to be
# originating from different locations due to the linking of parent directories.


# Prepare the file and links for the test
def _prepare_parent_directory_link_example(tmp_path):
    # Create data directory
    data_path = tmp_path / "data"

    # Create directory containing the actual files
    original_dir = data_path / 'original' / 'mydata' / 'v1.0.0'
    original_dir.mkdir(parents=True)

    # Create original file: file_a
    (original_dir / "file_a").write_text("secret", encoding='utf-8')

    # Create symbolic link: file_b -> file_a
    _create_symlink("file_a", original_dir / "file_b")

    # Create symbolic link: file_c -> file_b
    _create_symlink("file_b", original_dir / "file_c")

    # Create symbolic link: file_d -> file_c
    _create_symlink("file_c", original_dir / "file_d")

    # Create a symbolic link at the directory level
    linked_dir = data_path / "linked"
    linked_dir.mkdir(parents=True)

    _create_symlink(
        os.path.join('..', 'original', 'mydata', 'v1.0.0'),
        linked_dir / 'mydata-v1.0.0',
        target_is_directory=True,
    )

    return data_path


@pytest.mark.parametrize(
    "collected_files",
    [
        # Collect file_a from one directory and file_d from the other; this also causes missing intermediate links
        # (file_b, file_c).
        ['original/mydata/v1.0.0/file_a', 'linked/mydata-v1.0.0/file_d'],
        ['linked/mydata-v1.0.0/file_a', 'original/mydata/v1.0.0/file_d'],
    ],
    ids=['original', 'reversed'],
)
def test_symlinks__collect_chained_links_from_linked_directories(tmp_path, pyi_builder, collected_files):
    data_path = _prepare_parent_directory_link_example(tmp_path)
    pyi_builder.test_source(
        """
        import sys
        import os

        # file_a should be a regular file
        file_a = os.path.join(sys._MEIPASS, "file_a")
        assert os.path.isfile(file_a), "file_a does not exist!"
        assert not os.path.islink(file_a), "file_a is a symbolic link, but should not be!"

        # file_d should be a symlink pointing to file_a
        file_d = os.path.join(sys._MEIPASS, "file_d")
        assert os.path.isfile(file_d), "file_d does not exist!"
        assert os.path.islink(file_d), "file_d is not a symbolic link!"
        assert os.readlink(file_d) == "file_a", "file_d does not point to file_a!"
        """,
        pyi_args=list(_collect_data(data_path, collected_files))
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2014, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller import compat

# During python 3.14 development cycle, ctypes struct/union layout logic has been moved from `_ctypes` extension into
# Python, i.e., `ctypes._layout` module: https://github.com/python/cpython/pull/123352
# Since this module is referenced only from the `_ctypes` extension, it needs to be added to hidden imports, at least on
# Windows and macOS.
if compat.is_py314:
    hiddenimports = ['ctypes._layout']



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib.Tools import ccroot, ar, gcc
from waflib.Configure import conf


@conf
def find_clang(conf):
    cc = conf.find_program('clang', var='CC')
    conf.get_cc_version(cc, clang=True)
    conf.env.CC_NAME = 'clang'


def configure(conf):
    conf.find_clang()
    conf.find_program(['llvm-ar', 'ar'], var='AR')
    conf.find_ar()
    conf.gcc_common_flags()
    conf.gcc_modifier_platform()
    conf.cc_load_tools()
    conf.cc_add_flags()
    conf.link_add_flags()



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, re, sys, shutil
from waflib import Utils, Errors

exclude_regs = '''
**/*~
**/#*#
**/.#*
**/%*%
**/._*
**/*.swp
**/CVS
**/CVS/**
**/.cvsignore
**/SCCS
**/SCCS/**
**/vssver.scc
**/.svn
**/.svn/**
**/BitKeeper
**/.git
**/.git/**
**/.gitignore
**/.bzr
**/.bzrignore
**/.bzr/**
**/.hg
**/.hg/**
**/_MTN
**/_MTN/**
**/.arch-ids
**/{arch}
**/_darcs
**/_darcs/**
**/.intlcache
**/.DS_Store'''


def ant_matcher(s, ignorecase):
    reflags = re.I if ignorecase else 0
    ret = []
    for x in Utils.to_list(s):
        x = x.replace('\\', '/').replace('//', '/')
        if x.endswith('/'):
            x += '**'
        accu = []
        for k in x.split('/'):
            if k == '**':
                accu.append(k)
            else:
                k = k.replace('.', '[.]').replace('*', '.*').replace('?', '.').replace('+', '\\+')
                k = '^%s$' % k
                try:
                    exp = re.compile(k, flags=reflags)
                except Exception as e:
                    raise Errors.WafError('Invalid pattern: %s' % k, e)
                else:
                    accu.append(exp)
        ret.append(accu)
    return ret


def ant_sub_filter(name, nn):
    ret = []
    for lst in nn:
        if not lst:
            pass
        elif lst[0] == '**':
            ret.append(lst)
            if len(lst) > 1:
                if lst[1].match(name):
                    ret.append(lst[2:])
            else:
                ret.append([])
        elif lst[0].match(name):
            ret.append(lst[1:])
    return ret


def ant_sub_matcher(name, pats):
    nacc = ant_sub_filter(name, pats[0])
    nrej = ant_sub_filter(name, pats[1])
    if [] in nrej:
        nacc = []
    return [nacc, nrej]


class Node(object):
    dict_class = dict
    __slots__ = ('name', 'parent', 'children', 'cache_abspath', 'cache_isdir')

    def __init__(self, name, parent):
        self.name = name
        self.parent = parent
        if parent:
            if name in parent.children:
                raise Errors.WafError('node %s exists in the parent files %r already' % (name, parent))
            parent.children[name] = self

    def __setstate__(self, data):
        self.name = data[0]
        self.parent = data[1]
        if data[2] is not None:
            self.children = self.dict_class(data[2])

    def __getstate__(self):
        return (self.name, self.parent, getattr(self, 'children', None))

    def __str__(self):
        return self.abspath()

    def __repr__(self):
        return self.abspath()

    def __copy__(self):
        raise Errors.WafError('nodes are not supposed to be copied')

    def read(self, flags='r', encoding='latin-1'):
        return Utils.readf(self.abspath(), flags, encoding)

    def write(self, data, flags='w', encoding='latin-1'):
        Utils.writef(self.abspath(), data, flags, encoding)

    def read_json(self, convert=True, encoding='utf-8'):
        import json
        object_pairs_hook = None
        if convert and sys.hexversion < 0x3000000:
            try:
                _type = unicode
            except NameError:
                _type = str

            def convert(value):
                if isinstance(value, list):
                    return [convert(element) for element in value]
                elif isinstance(value, _type):
                    return str(value)
                else:
                    return value

            def object_pairs(pairs):
                return dict((str(pair[0]), convert(pair[1])) for pair in pairs)

            object_pairs_hook = object_pairs
        return json.loads(self.read(encoding=encoding), object_pairs_hook=object_pairs_hook)

    def write_json(self, data, pretty=True):
        import json
        indent = 2
        separators = (',', ': ')
        sort_keys = pretty
        newline = os.linesep
        if not pretty:
            indent = None
            separators = (',', ':')
            newline = ''
        output = json.dumps(data, indent=indent, separators=separators, sort_keys=sort_keys) + newline
        self.write(output, encoding='utf-8')

    def exists(self):
        return os.path.exists(self.abspath())

    def isdir(self):
        return os.path.isdir(self.abspath())

    def chmod(self, val):
        os.chmod(self.abspath(), val)

    def delete(self, evict=True):
        try:
            try:
                if os.path.isdir(self.abspath()):
                    shutil.rmtree(self.abspath())
                else:
                    os.remove(self.abspath())
            except OSError:
                if os.path.exists(self.abspath()):
                    raise
        finally:
            if evict:
                self.evict()

    def evict(self):
        del self.parent.children[self.name]

    def suffix(self):
        k = max(0, self.name.rfind('.'))
        return self.name[k:]

    def height(self):
        d = self
        val = -1
        while d:
            d = d.parent
            val += 1
        return val

    def listdir(self):
        lst = Utils.listdir(self.abspath())
        lst.sort()
        return lst

    def mkdir(self):
        if self.isdir():
            return
        try:
            self.parent.mkdir()
        except OSError:
            pass
        if self.name:
            try:
                os.makedirs(self.abspath())
            except OSError:
                pass
            if not self.isdir():
                raise Errors.WafError('Could not create the directory %r' % self)
            try:
                self.children
            except AttributeError:
                self.children = self.dict_class()

    def find_node(self, lst):
        if isinstance(lst, str):
            lst = [x for x in Utils.split_path(lst) if x and x != '.']
        if lst and lst[0].startswith('\\\\') and not self.parent:
            node = self.ctx.root.make_node(lst[0])
            node.cache_isdir = True
            return node.find_node(lst[1:])
        cur = self
        for x in lst:
            if x == '..':
                cur = cur.parent or cur
                continue
            try:
                ch = cur.children
            except AttributeError:
                cur.children = self.dict_class()
            else:
                try:
                    cur = ch[x]
                    continue
                except KeyError:
                    pass
            cur = self.__class__(x, cur)
            if not cur.exists():
                cur.evict()
                return None
        if not cur.exists():
            cur.evict()
            return None
        return cur

    def make_node(self, lst):
        if isinstance(lst, str):
            lst = [x for x in Utils.split_path(lst) if x and x != '.']
        cur = self
        for x in lst:
            if x == '..':
                cur = cur.parent or cur
                continue
            try:
                cur = cur.children[x]
            except AttributeError:
                cur.children = self.dict_class()
            except KeyError:
                pass
            else:
                continue
            cur = self.__class__(x, cur)
        return cur

    def search_node(self, lst):
        if isinstance(lst, str):
            lst = [x for x in Utils.split_path(lst) if x and x != '.']
        cur = self
        for x in lst:
            if x == '..':
                cur = cur.parent or cur
            else:
                try:
                    cur = cur.children[x]
                except (AttributeError, KeyError):
                    return None
        return cur

    def path_from(self, node):
        c1 = self
        c2 = node
        c1h = c1.height()
        c2h = c2.height()
        lst = []
        up = 0
        while c1h > c2h:
            lst.append(c1.name)
            c1 = c1.parent
            c1h -= 1
        while c2h > c1h:
            up += 1
            c2 = c2.parent
            c2h -= 1
        while not c1 is c2:
            lst.append(c1.name)
            up += 1
            c1 = c1.parent
            c2 = c2.parent
        if c1.parent:
            lst.extend(['..'] * up)
            lst.reverse()
            return os.sep.join(lst) or '.'
        else:
            return self.abspath()

    def abspath(self):
        try:
            return self.cache_abspath
        except AttributeError:
            pass
        if not self.parent:
            val = os.sep
        elif not self.parent.name:
            val = os.sep + self.name
        else:
            val = self.parent.abspath() + os.sep + self.name
        self.cache_abspath = val
        return val

    if Utils.is_win32:

        def abspath(self):
            try:
                return self.cache_abspath
            except AttributeError:
                pass
            if not self.parent:
                val = ''
            elif not self.parent.name:
                val = self.name + os.sep
            else:
                val = self.parent.abspath().rstrip(os.sep) + os.sep + self.name
            self.cache_abspath = val
            return val

    def is_child_of(self, node):
        p = self
        diff = self.height() - node.height()
        while diff > 0:
            diff -= 1
            p = p.parent
        return p is node

    def ant_iter(self, accept=None, maxdepth=25, pats=[], dir=False, src=True, remove=True, quiet=False):
        dircont = self.listdir()
        try:
            lst = set(self.children.keys())
        except AttributeError:
            self.children = self.dict_class()
        else:
            if remove:
                for x in lst - set(dircont):
                    self.children[x].evict()
        for name in dircont:
            npats = accept(name, pats)
            if npats and npats[0]:
                accepted = [] in npats[0]
                node = self.make_node([name])
                isdir = node.isdir()
                if accepted:
                    if isdir:
                        if dir:
                            yield node
                    elif src:
                        yield node
                if isdir:
                    node.cache_isdir = True
                    if maxdepth:
                        for k in node.ant_iter(
                            accept=accept,
                            maxdepth=maxdepth - 1,
                            pats=npats,
                            dir=dir,
                            src=src,
                            remove=remove,
                            quiet=quiet
                        ):
                            yield k

    def ant_glob(self, *k, **kw):
        src = kw.get('src', True)
        dir = kw.get('dir')
        excl = kw.get('excl', exclude_regs)
        incl = k and k[0] or kw.get('incl', '**')
        remove = kw.get('remove', True)
        maxdepth = kw.get('maxdepth', 25)
        ignorecase = kw.get('ignorecase', False)
        quiet = kw.get('quiet', False)
        pats = (ant_matcher(incl, ignorecase), ant_matcher(excl, ignorecase))
        if kw.get('generator'):
            return Utils.lazy_generator(self.ant_iter, (ant_sub_matcher, maxdepth, pats, dir, src, remove, quiet))
        it = self.ant_iter(ant_sub_matcher, maxdepth, pats, dir, src, remove, quiet)
        if kw.get('flat'):
            return ' '.join(x.path_from(self) for x in it)
        return list(it)

    def is_src(self):
        cur = self
        x = self.ctx.srcnode
        y = self.ctx.bldnode
        while cur.parent:
            if cur is y:
                return False
            if cur is x:
                return True
            cur = cur.parent
        return False

    def is_bld(self):
        cur = self
        y = self.ctx.bldnode
        while cur.parent:
            if cur is y:
                return True
            cur = cur.parent
        return False

    def get_src(self):
        cur = self
        x = self.ctx.srcnode
        y = self.ctx.bldnode
        lst = []
        while cur.parent:
            if cur is y:
                lst.reverse()
                return x.make_node(lst)
            if cur is x:
                return self
            lst.append(cur.name)
            cur = cur.parent
        return self

    def get_bld(self):
        cur = self
        x = self.ctx.srcnode
        y = self.ctx.bldnode
        lst = []
        while cur.parent:
            if cur is y:
                return self
            if cur is x:
                lst.reverse()
                return self.ctx.bldnode.make_node(lst)
            lst.append(cur.name)
            cur = cur.parent
        lst.reverse()
        if lst and Utils.is_win32 and len(lst[0]) == 2 and lst[0].endswith(':'):
            lst[0] = lst[0][0]
        return self.ctx.bldnode.make_node(['__root__'] + lst)

    def find_resource(self, lst):
        if isinstance(lst, str):
            lst = [x for x in Utils.split_path(lst) if x and x != '.']
        node = self.get_bld().search_node(lst)
        if not node:
            node = self.get_src().find_node(lst)
        if node and node.isdir():
            return None
        return node

    def find_or_declare(self, lst):
        if isinstance(lst, str) and os.path.isabs(lst):
            node = self.ctx.root.make_node(lst)
        else:
            node = self.get_bld().make_node(lst)
        node.parent.mkdir()
        return node

    def find_dir(self, lst):
        if isinstance(lst, str):
            lst = [x for x in Utils.split_path(lst) if x and x != '.']
        node = self.find_node(lst)
        if node and not node.isdir():
            return None
        return node

    def change_ext(self, ext, ext_in=None):
        name = self.name
        if ext_in is None:
            k = name.rfind('.')
            if k >= 0:
                name = name[:k] + ext
            else:
                name = name + ext
        else:
            name = name[:-len(ext_in)] + ext
        return self.parent.find_or_declare([name])

    def bldpath(self):
        return self.path_from(self.ctx.bldnode)

    def srcpath(self):
        return self.path_from(self.ctx.srcnode)

    def relpath(self):
        cur = self
        x = self.ctx.bldnode
        while cur.parent:
            if cur is x:
                return self.bldpath()
            cur = cur.parent
        return self.srcpath()

    def bld_dir(self):
        return self.parent.bldpath()

    def h_file(self):
        return Utils.h_file(self.abspath())

    def get_bld_sig(self):
        try:
            cache = self.ctx.cache_sig
        except AttributeError:
            cache = self.ctx.cache_sig = {}
        try:
            ret = cache[self]
        except KeyError:
            p = self.abspath()
            try:
                ret = cache[self] = self.h_file()
            except EnvironmentError:
                if self.isdir():
                    st = os.stat(p)
                    ret = cache[self] = Utils.h_list([p, st.st_ino, st.st_mode])
                    return ret
                raise
        return ret


pickle_lock = Utils.threading.Lock()


class Nod3(Node):
    pass



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import re, os, sys, shlex
from waflib.Configure import conf
from waflib.TaskGen import feature, before_method

FC_FRAGMENT = '        program main\n        end     program main\n'
FC_FRAGMENT2 = '        PROGRAM MAIN\n        END\n'


@conf
def fc_flags(conf):
    v = conf.env
    v.FC_SRC_F = []
    v.FC_TGT_F = ['-c', '-o']
    v.FCINCPATH_ST = '-I%s'
    v.FCDEFINES_ST = '-D%s'
    if not v.LINK_FC:
        v.LINK_FC = v.FC
    v.FCLNK_SRC_F = []
    v.FCLNK_TGT_F = ['-o']
    v.FCFLAGS_fcshlib = ['-fpic']
    v.LINKFLAGS_fcshlib = ['-shared']
    v.fcshlib_PATTERN = 'lib%s.so'
    v.fcstlib_PATTERN = 'lib%s.a'
    v.FCLIB_ST = '-l%s'
    v.FCLIBPATH_ST = '-L%s'
    v.FCSTLIB_ST = '-l%s'
    v.FCSTLIBPATH_ST = '-L%s'
    v.FCSTLIB_MARKER = '-Wl,-Bstatic'
    v.FCSHLIB_MARKER = '-Wl,-Bdynamic'
    v.SONAME_ST = '-Wl,-h,%s'


@conf
def fc_add_flags(conf):
    conf.add_os_flags('FCPPFLAGS', dup=False)
    conf.add_os_flags('FCFLAGS', dup=False)
    conf.add_os_flags('LINKFLAGS', dup=False)
    conf.add_os_flags('LDFLAGS', dup=False)


@conf
def check_fortran(self, *k, **kw):
    self.check_cc(
        fragment=FC_FRAGMENT, compile_filename='test.f', features='fc fcprogram', msg='Compiling a simple fortran app'
    )


@conf
def check_fc(self, *k, **kw):
    kw['compiler'] = 'fc'
    if not 'compile_mode' in kw:
        kw['compile_mode'] = 'fc'
    if not 'type' in kw:
        kw['type'] = 'fcprogram'
    if not 'compile_filename' in kw:
        kw['compile_filename'] = 'test.f90'
    if not 'code' in kw:
        kw['code'] = FC_FRAGMENT
    return self.check(*k, **kw)


@conf
def fortran_modifier_darwin(conf):
    v = conf.env
    v.FCFLAGS_fcshlib = ['-fPIC']
    v.LINKFLAGS_fcshlib = ['-dynamiclib']
    v.fcshlib_PATTERN = 'lib%s.dylib'
    v.FRAMEWORKPATH_ST = '-F%s'
    v.FRAMEWORK_ST = ['-framework']
    v.LINKFLAGS_fcstlib = []
    v.FCSHLIB_MARKER = ''
    v.FCSTLIB_MARKER = ''
    v.SONAME_ST = ''


@conf
def fortran_modifier_win32(conf):
    v = conf.env
    v.fcprogram_PATTERN = v.fcprogram_test_PATTERN = '%s.exe'
    v.fcshlib_PATTERN = '%s.dll'
    v.implib_PATTERN = '%s.dll.a'
    v.IMPLIB_ST = '-Wl,--out-implib,%s'
    v.FCFLAGS_fcshlib = []
    v.append_value('LINKFLAGS', ['-Wl,--enable-auto-import'])


@conf
def fortran_modifier_cygwin(conf):
    fortran_modifier_win32(conf)
    v = conf.env
    v.fcshlib_PATTERN = 'cyg%s.dll'
    v.append_value('LINKFLAGS_fcshlib', ['-Wl,--enable-auto-image-base'])
    v.FCFLAGS_fcshlib = []


@conf
def check_fortran_dummy_main(self, *k, **kw):
    if not self.env.CC:
        self.fatal('A c compiler is required for check_fortran_dummy_main')
    lst = ['MAIN__', '__MAIN', '_MAIN', 'MAIN_', 'MAIN']
    lst.extend([m.lower() for m in lst])
    lst.append('')
    self.start_msg('Detecting whether we need a dummy main')
    for main in lst:
        kw['fortran_main'] = main
        try:
            self.check_cc(
                fragment='int %s() { return 0; }\n' % (main or 'test'), features='c fcprogram', mandatory=True
            )
            if not main:
                self.env.FC_MAIN = -1
                self.end_msg('no')
            else:
                self.env.FC_MAIN = main
                self.end_msg('yes %s' % main)
            break
        except self.errors.ConfigurationError:
            pass
    else:
        self.end_msg('not found')
        self.fatal('could not detect whether fortran requires a dummy main, see the config.log')


GCC_DRIVER_LINE = re.compile('^Driving:')
POSIX_STATIC_EXT = re.compile(r'\S+\.a')
POSIX_LIB_FLAGS = re.compile(r'-l\S+')


@conf
def is_link_verbose(self, txt):
    assert isinstance(txt, str)
    for line in txt.splitlines():
        if not GCC_DRIVER_LINE.search(line):
            if POSIX_STATIC_EXT.search(line) or POSIX_LIB_FLAGS.search(line):
                return True
    return False


@conf
def check_fortran_verbose_flag(self, *k, **kw):
    self.start_msg('fortran link verbose flag')
    for x in ('-v', '--verbose', '-verbose', '-V'):
        try:
            self.check_cc(
                features='fc fcprogram_test',
                fragment=FC_FRAGMENT2,
                compile_filename='test.f',
                linkflags=[x],
                mandatory=True
            )
        except self.errors.ConfigurationError:
            pass
        else:
            if self.is_link_verbose(self.test_bld.err) or self.is_link_verbose(self.test_bld.out):
                self.end_msg(x)
                break
    else:
        self.end_msg('failure')
        self.fatal('Could not obtain the fortran link verbose flag (see config.log)')
    self.env.FC_VERBOSE_FLAG = x
    return x


LINKFLAGS_IGNORED = [r'-lang*', r'-lcrt[a-zA-Z0-9\.]*\.o', r'-lc$', r'-lSystem', r'-libmil', r'-LIST:*', r'-LNO:*']
if os.name == 'nt':
    LINKFLAGS_IGNORED.extend([
        r'-lfrt*', r'-luser32', r'-lkernel32', r'-ladvapi32', r'-lmsvcrt', r'-lshell32', r'-lmingw', r'-lmoldname'
    ])
else:
    LINKFLAGS_IGNORED.append(r'-lgcc*')
RLINKFLAGS_IGNORED = [re.compile(f) for f in LINKFLAGS_IGNORED]


def _match_ignore(line):
    for i in RLINKFLAGS_IGNORED:
        if i.match(line):
            return True
    return False


def parse_fortran_link(lines):
    final_flags = []
    for line in lines:
        if not GCC_DRIVER_LINE.match(line):
            _parse_flink_line(line, final_flags)
    return final_flags


SPACE_OPTS = re.compile('^-[LRuYz]$')
NOSPACE_OPTS = re.compile('^-[RL]')


def _parse_flink_token(lexer, token, tmp_flags):
    if _match_ignore(token):
        pass
    elif token.startswith('-lkernel32') and sys.platform == 'cygwin':
        tmp_flags.append(token)
    elif SPACE_OPTS.match(token):
        t = lexer.get_token()
        if t.startswith('P,'):
            t = t[2:]
        for opt in t.split(os.pathsep):
            tmp_flags.append('-L%s' % opt)
    elif NOSPACE_OPTS.match(token):
        tmp_flags.append(token)
    elif POSIX_LIB_FLAGS.match(token):
        tmp_flags.append(token)
    else:
        pass
    t = lexer.get_token()
    return t


def _parse_flink_line(line, final_flags):
    lexer = shlex.shlex(line, posix=True)
    lexer.whitespace_split = True
    t = lexer.get_token()
    tmp_flags = []
    while t:
        t = _parse_flink_token(lexer, t, tmp_flags)
    final_flags.extend(tmp_flags)
    return final_flags


@conf
def check_fortran_clib(self, autoadd=True, *k, **kw):
    if not self.env.FC_VERBOSE_FLAG:
        self.fatal('env.FC_VERBOSE_FLAG is not set: execute check_fortran_verbose_flag?')
    self.start_msg('Getting fortran runtime link flags')
    try:
        self.check_cc(
            fragment=FC_FRAGMENT2,
            compile_filename='test.f',
            features='fc fcprogram_test',
            linkflags=[self.env.FC_VERBOSE_FLAG]
        )
    except Exception:
        self.end_msg(False)
        if kw.get('mandatory', True):
            conf.fatal('Could not find the c library flags')
    else:
        out = self.test_bld.err
        flags = parse_fortran_link(out.splitlines())
        self.end_msg('ok (%s)' % ' '.join(flags))
        self.env.LINKFLAGS_CLIB = flags
        return flags
    return []


def getoutput(conf, cmd, stdin=False):
    from waflib import Errors
    if conf.env.env:
        env = conf.env.env
    else:
        env = dict(os.environ)
        env['LANG'] = 'C'
    input = stdin and '\n'.encode() or None
    try:
        out, err = conf.cmd_and_log(cmd, env=env, output=0, input=input)
    except Errors.WafError as e:
        if not (hasattr(e, 'stderr') and hasattr(e, 'stdout')):
            raise e
        else:
            out = e.stdout
            err = e.stderr
    except Exception:
        conf.fatal('could not determine the compiler version %r' % cmd)
    return (out, err)


ROUTINES_CODE = """\
      subroutine foobar()
      return
      end
      subroutine foo_bar()
      return
      end
"""
MAIN_CODE = """
void %(dummy_func_nounder)s(void);
void %(dummy_func_under)s(void);
int %(main_func_name)s() {
  %(dummy_func_nounder)s();
  %(dummy_func_under)s();
  return 0;
}
"""


@feature('link_main_routines_func')
@before_method('process_source')
def link_main_routines_tg_method(self):
    def write_test_file(task):
        task.outputs[0].write(task.generator.code)

    bld = self.bld
    bld(rule=write_test_file, target='main.c', code=MAIN_CODE % self.__dict__)
    bld(rule=write_test_file, target='test.f', code=ROUTINES_CODE)
    bld(features='fc fcstlib', source='test.f', target='test')
    bld(features='c fcprogram', source='main.c', target='app', use='test')


def mangling_schemes():
    for u in ('_', ''):
        for du in ('', '_'):
            for c in ("lower", "upper"):
                yield (u, du, c)


def mangle_name(u, du, c, name):
    return getattr(name, c)() + u + (name.find('_') != -1 and du or '')


@conf
def check_fortran_mangling(self, *k, **kw):
    if not self.env.CC:
        self.fatal('A c compiler is required for link_main_routines')
    if not self.env.FC:
        self.fatal('A fortran compiler is required for link_main_routines')
    if not self.env.FC_MAIN:
        self.fatal('Checking for mangling requires self.env.FC_MAIN (execute "check_fortran_dummy_main" first?)')
    self.start_msg('Getting fortran mangling scheme')
    for (u, du, c) in mangling_schemes():
        try:
            self.check_cc(
                compile_filename=[],
                features='link_main_routines_func',
                msg='nomsg',
                errmsg='nomsg',
                dummy_func_nounder=mangle_name(u, du, c, 'foobar'),
                dummy_func_under=mangle_name(u, du, c, 'foo_bar'),
                main_func_name=self.env.FC_MAIN
            )
        except self.errors.ConfigurationError:
            pass
        else:
            self.end_msg("ok ('%s', '%s', '%s-case')" % (u, du, c))
            self.env.FORTRAN_MANGLING = (u, du, c)
            break
    else:
        self.end_msg(False)
        self.fatal('mangler not found')
    return (u, du, c)


@feature('pyext')
@before_method('propagate_uselib_vars', 'apply_link')
def set_lib_pat(self):
    self.env.fcshlib_PATTERN = self.env.pyext_PATTERN


@conf
def detect_openmp(self):
    for x in ('-fopenmp', '-openmp', '-mp', '-xopenmp', '-omp', '-qsmp=omp'):
        try:
            self.check_fc(
                msg='Checking for OpenMP flag %s' % x,
                fragment='program main\n  call omp_get_num_threads()\nend program main',
                fcflags=x,
                linkflags=x,
                uselib_store='OPENMP'
            )
        except self.errors.ConfigurationError:
            pass
        else:
            break
    else:
        self.fatal('Could not find OpenMP')


@conf
def check_gfortran_o_space(self):
    if self.env.FC_NAME != 'GFORTRAN' or int(self.env.FC_VERSION[0]) > 4:
        return
    self.env.stash()
    self.env.FCLNK_TGT_F = ['-o', '']
    try:
        self.check_fc(
            msg='Checking if the -o link must be split from arguments', fragment=FC_FRAGMENT, features='fc fcshlib'
        )
    except self.errors.ConfigurationError:
        self.env.revert()
    else:
        self.env.commit()



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import pathlib

import pytest

from PyInstaller.compat import exec_python_rc
from PyInstaller.utils.tests import importable, xfail, onedir_only, onefile_only

# Directory with testing modules used in some tests.
_MODULES_DIR = pathlib.Path(__file__).parent / 'modules'


# Tests for pkgutil.get_data().
def test_pkgutil_get_data(pyi_builder):
    add_data_arg = f"{_MODULES_DIR / 'pkg3' / 'sample-data.txt'}:pkg3"
    pyi_builder.test_source(
        """
            import pkgutil
            import pkg3  # Serves as a hiddenimport

            expected_data = b'This is data text for testing the packaging module data.'

            data = pkgutil.get_data('pkg3', 'sample-data.txt')
            print("Read data: {data!r}")
            if not data:
                raise SystemExit('Error: Could not read data with pkgutil.get_data().')

            if data.strip() != expected_data:
                raise SystemExit('Error: Read data does not match expected data!')
        """,
        pyi_args=['--add-data', add_data_arg],
    )


@xfail(reason='Our import mechanism returns the wrong loader-class for __main__.')
def test_pkgutil_get_data__main__(pyi_builder):
    add_data_arg = f"{_MODULES_DIR / 'pkg3' / 'sample-data.txt'}:pkg3"
    pyi_builder.test_source(
        """
        import pkgutil

        expected_data = b'This is data text for testing the packaging module data.'

        data = pkgutil.get_data('__main__', 'pkg3/sample-data.txt')
        if not data:
            raise SystemExit('Error: Could not read data with pkgutil.get_data().')

        if data.strip() != expected_data:
            raise SystemExit('Error: Read data does not match expected data!')
        """,
        pyi_args=['--add-data', add_data_arg],
    )


# Tests for pkgutil.iter_modules(). The test attempts to list contents of a package in both unfrozen and frozen version,
# and compares the obtained lists.
#
# We test three scenarios; a pure-python top-level package (using json package from stdlib), a pure-python sub-package
# (using xml.dom package from stdlib), and a package with binary extensions (using psutil).
#
# The extensions are present on filesystem as-is, and are therefore handled by python's FileFinder. The collected .pyc
# modules, however, are embedded in PYZ archive, and are not visible to standard python's finders/loaders. The exception
# to that is noarchive mode, where .pyc modules are not collected into archive; as they are present on filesystem as-is,
# they are again handled directly by python's FileFinder. Therefore, each test is performed both in archive and in
# noarchive mode, to cover both cases.


# Read the output file produced by test script. Each line consists of two elements separated by semi-colon:
# name;ispackage
def _read_results_file(filename):
    output = []
    with open(filename, 'r', encoding='utf-8') as fp:
        for line in fp:
            tokens = line.split(';')
            assert len(tokens) == 2
            output.append((tokens[0], int(tokens[1])))
    # Sort the results, so we can compare them
    return sorted(output)


@pytest.mark.parametrize(
    'package',
    [
        'json',  # pure python package (stdlib)
        'xml.dom',  # sub-package (stdlib)
        'psutil',  # package with extensions (3rd party)
    ]
)
@pytest.mark.parametrize('archive', ['archive', 'noarchive'])
def test_pkgutil_iter_modules(package, script_dir, tmp_path, pyi_builder, archive, resolve_pkg_path=False):
    # Ensure package is available
    if not importable(package.split(".")[0]):
        pytest.skip(f"Needs {package}")

    # Full path to test script
    test_script = script_dir / 'pyi_pkgutil_iter_modules.py'

    # Run unfrozen test script
    out_unfrozen = tmp_path / 'output-unfrozen.txt'
    rc = exec_python_rc(str(test_script), package, '--output-file', str(out_unfrozen))
    assert rc == 0
    # Read results
    results_unfrozen = _read_results_file(out_unfrozen)

    # Run frozen script
    out_frozen = tmp_path / 'output-frozen.txt'
    debug_args = ['--debug', 'noarchive'] if archive == 'noarchive' else []
    pyi_builder.test_script(
        test_script,
        pyi_args=[
            # ensure everything is collected
            '--collect-submodules', package,
            # enable/disable noarchive
            *debug_args,
        ],
        app_args=[package, '--output-file', str(out_frozen)] + (['--resolve-pkg-path'] if resolve_pkg_path else [])
    )  # yapf: disable
    # Read results
    results_frozen = _read_results_file(out_frozen)

    # Compare
    assert results_unfrozen == results_frozen


# Repeat test_pkgutil_iter_modules() test with package path resolving enabled. In this mode, the test script fully
# resolves the package path before passing it to pkgutil.iter_modules(), reproducing the scenario of #6537 on macOS:
# the temporary directory used by onefile builds is placed in /var, which is symbolic link to /private/var. Therefore,
# the resolved package path (/private/var/...) in the frozen application may differ from non-resolved one (/var/...),
# and our pkgutil.iter_modules() therefore needs to explicitly resolve the given paths and the sys._MEIPASS prefix to
# ensure proper matching.
# The test is applicable only to macOS in onefile mode.
@pytest.mark.darwin
@onefile_only
def test_pkgutil_iter_modules_resolve_pkg_path(script_dir, tmp_path, pyi_builder):
    # A single combination (altgraph package, archive mode) is enough to check for proper symlink handling.
    test_pkgutil_iter_modules('json', script_dir, tmp_path, pyi_builder, archive=True, resolve_pkg_path=True)


# Additional test for macOS .app bundles and packages that contain data files. See #7884. In generated .app bundles,
# _MEIPASS points to `Contents/Frameworks`, while the data files are collected into `Contents/Resources` directory. If
# a package contains only data files, the whole package directory is collected into `Contents/Resources`, and a symbolic
# link to package's directory is made in `Contents/Frameworks`. Our `pkgutil.iter_modules` implementation needs to
# account for this when validating the package path prefix; i.e., that attempting to resolve
# `Contents/Frameworks/mypackage` will result in `Contents/Resource/mypackage` due to symbolic link, and thus the prefix
# will not directly match _MEIPASS anymore.
#
# This issue affects packages with only data files; if the package has no data or binary files, then the package
# directory does not exist on filesystem and the resolution attempt leaves it unchanged. If the package contains both
# data and binary files, the directory is created in both Contents/Frameworks and Contents/Resources, and the contents
# are cross-linked between them on file level.
@pytest.mark.darwin
@onedir_only
def test_pkgutil_iter_modules_macos_app_bundle(script_dir, tmp_path, pyi_builder, monkeypatch):
    pathex = _MODULES_DIR / 'pyi_pkgutil_itermodules' / 'package'
    hooks_dir = _MODULES_DIR / 'pyi_pkgutil_itermodules' / 'hooks'
    package = 'mypackage'

    # Full path to test script
    test_script = script_dir / 'pyi_pkgutil_iter_modules.py'

    # Run unfrozen test script
    env = os.environ.copy()
    if 'PYTHONPATH' in env:
        pathex = os.pathsep.join([str(pathex), env['PYTHONPATH']])
    env['PYTHONPATH'] = str(pathex)
    out_unfrozen = tmp_path / 'output-unfrozen.txt'
    rc = exec_python_rc(str(test_script), package, '--output-file', str(out_unfrozen), env=env)
    assert rc == 0
    # Read results
    results_unfrozen = _read_results_file(out_unfrozen)

    # Freeze the test program
    # This also runs both executables (POSIX build and .app bundle) with same arguments, so we have no way of separating
    # the output file. Therefore, we will manually re-run the executables ourselves.
    pyi_builder.test_script(
        test_script,
        pyi_args=[
            '--paths', str(pathex),
            '--hiddenimport', package,
            '--additional-hooks-dir', str(hooks_dir),
            '--windowed',  # enable .app bundle
        ],
        app_args=[package],
    )  # yapf: disable

    # Run each executable and verify its output
    executables = pyi_builder._find_executables('pyi_pkgutil_iter_modules')
    assert executables
    for idx, exe in enumerate(executables):
        out_frozen = tmp_path / f"output-frozen-{idx}.txt"
        rc = pyi_builder._run_executable(
            exe,
            args=[package, '--output-file', str(out_frozen)],
            run_from_path=False,
            runtime=None,
        )
        assert rc == 0
        results_frozen = _read_results_file(out_frozen)
        print("RESULTS", results_frozen, "\n\n")

        assert results_unfrozen == results_frozen


# Explicitly test that in macOS .app bundles, modules can be iterated regardless of whether the given search path is
# anchored in the Contents/Frameworks directory (the "true" sys._MEIPASS) or the Contents/Resources directory.
# A real-world example would involve a package that is partially collected into PYZ, but its `__init__` module is
# collected as source .py file only. Thus, the search path derived from fully resolved __file__ attribute would end up
# pointing to Resources directory instead of Frameworks one. For the purpose of the test, however, we simply modify the
# search path ourselves.
#
# This is more explicit version of test_pkgutil_iter_modules_macos_app_bundle, just in case.
@pytest.mark.darwin
@onedir_only
def test_pkgutil_iter_modules_macos_app_bundle_alternative_search_path(pyi_builder):
    pyi_builder.test_source(
        """
        import os
        import sys
        import pkgutil
        import json  # Our test package

        # Check that we are running in .app bundle mode. If not, exit.
        print("sys._MEIPASS:", sys._MEIPASS)
        if not sys._MEIPASS.endswith("Contents/Frameworks"):
            print("Not running as .app bundle.")
            sys.exit(0)

        alternative_top_level_dir = os.path.join(os.path.dirname(sys._MEIPASS), 'Resources')

        def _compare_path_contents(true_path, alternative_path, prefix=""):
            # Iterate over modules using path anchored to "true" top-level directory (sys._MEIPASS).
            print(f"Iterating over modules in path anchored to true top-level directory - {true_path!r}:")
            modules1 = list(pkgutil.iter_modules([true_path], prefix))
            for entry in modules1:
                print(entry)
            print("")

            assert len(modules1) > 0, "Modules list is emtpy?!"

            # Iterate over modules using path anchored to "alternative" top-level directory.
            print(f"Iterating over modules in path anchored to alternative top-level directory {alternative_path!r}:")
            modules2 = list(pkgutil.iter_modules([alternative_path], prefix))
            for entry in modules2:
                print(entry)
            print("")

            # We can compare only .name and .ispkg, because .module_finder might be per-path instance.
            def _to_comparable_list(modules):
                return sorted([(module.name, module.ispkg) for module in modules])

            assert _to_comparable_list(modules1) == _to_comparable_list(modules2), "Lists of modules do not match!"
            print("OK!")

        # First, run comparison on top-level application directory
        print("Running test for top-level application directory...")
        _compare_path_contents(
            sys._MEIPASS,
            alternative_top_level_dir,
        )

        # Repeat for the 'json' package
        print("Running test for 'json' package...")
        _compare_path_contents(
            os.path.join(sys._MEIPASS, 'json'),
            os.path.join(alternative_top_level_dir, 'json'),
            prefix='json.',
        )
        """,
        pyi_args=['--collect-submodules', 'json', '--windowed']
    )


# Two tests that reproduce the situation from #8191. In the first test, `pkgutil.iter_modules()` is called on a path
# that corresponds to a module instead of the package. In the second test, we add a sub-directory component to the path
# that corresponds to a module. Both cases should be handled gracefully by our `iter_modules` override.
def test_pkgutil_iter_modules_with_module_path(pyi_builder):
    pyi_builder.test_source(
        """
        import os
        import pkgutil
        import json.encoder  # Our test module

        # Path to iterate over; sys._MEIPASS/json/encoder
        search_path, _ = os.path.splitext(json.encoder.__file__)

        # pkgutil.iter_modules()
        print("Search path:", search_path)
        entries = list(pkgutil.iter_modules([search_path]))
        print("Entries:", entries)
        assert len(entries) == 0, "Expected no entries!"
        """
    )


def test_pkgutil_iter_modules_with_module_path_subdir(pyi_builder):
    pyi_builder.test_source(
        """
        import os
        import pkgutil
        import json.encoder  # Our test module

        # Path to iterate over; sys._MEIPASS/json/encoder/nonexistent
        search_path, _ = os.path.splitext(json.encoder.__file__)
        search_path = os.path.join(search_path, 'nonexistent')

        # pkgutil.iter_modules()
        print("Search path:", search_path)
        entries = list(pkgutil.iter_modules([search_path]))
        print("Entries:", entries)
        assert len(entries) == 0, "Expected no entries!"
        """
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# Test that the Python's `site` module is disabled and Python is not searching for any user-specific site directories;
# in frozen application, this is effectively checking that PyInstaller's bootloader sets the `site_import` flag in the
# `PyConfig` structure is set to 0 (equivalent of passing -S to python interpreter executable).

import sys

# With `python -S` option (and equivalent `PyConfig.site_import = 0`) set, Python should not import `site` module on
# the startup. Therefore, check that is not imported yet.
#
# NOTE: in the frozen application, the module might end up imported as a dependency of another module that is imported
# when PyInstaller's run-time hooks are imported. If the frozen test script fails at this point, it is best to check
# what run-time hooks were ran, and what imports were made from them; this test script is very simple, so it should
# not warrant imports of many modules with run-time hooks. However, a package like `setuptools` might end up being
# pulled in, for example, via reference to `distutils` (for which it provides replacement).
if 'site' in sys.modules:
    raise SystemExit('site module already imported')

import site

# Check that it really is disabled.
if not sys.flags.no_site:
    raise SystemExit('site module is enabled!')

# Default values of attributes from 'site' module when it is disabled.
# Under python 3, ENABLE_USER_SITE should be None.
if site.ENABLE_USER_SITE is not None:
    raise SystemExit(f'ENABLE_USER_SITE is {site.ENABLE_USER_SITE}, expected None!')

# Since we import `site` here in the test, this causes USER_SITE and USER_BASE to be initialized on Py2,
# so all we can do is confirm that the paths aren't in sys.path
if site.USER_SITE is not None:
    if site.USER_SITE in sys.path:
        raise SystemExit('USER_SITE found in sys.path!')

# This should never happen, USER_BASE is not a site-modules folder and is only used by distutils
# for installing module datas.
if site.USER_BASE is not None:
    if site.USER_BASE in sys.path:
        raise SystemExit('USER_BASE found in sys.path!')



#-----------------------------------------------------------------------------
# Copyright (c) 2024, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# This module conditionally imports PyQt5:
# https://github.com/pandas-dev/pandas/blob/95308514e1221200e4526dfaf248283f3d7ade06/pandas/io/clipboard/__init__.py#L578-L597
# Suppress this import to prevent PyQt5 from being accidentally pulled in; the actually relevant Qt bindings are
# determined by our hook for `qtpy` module, which contemporary versions of pandas mandate as part of `clipboard` and
# `all` extras:
# https://github.com/pandas-dev/pandas/blob/95308514e1221200e4526dfaf248283f3d7ade06/pyproject.toml#L86
# https://github.com/pandas-dev/pandas/blob/95308514e1221200e4526dfaf248283f3d7ade06/pyproject.toml#L115
excludedimports = ['PyQt5']



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller.depend.imphook import AdditionalFilesCache


def test_binaries_and_datas():
    datas = [('source', 'dest'), ('src', 'dst')]
    binaries = [('abc', 'def'), ('ghi', 'jkl')]
    modules = ['testmodule1', 'testmodule2']

    cache = AdditionalFilesCache()
    for modname in modules:
        cache.add(modname, binaries, datas)
        assert cache.datas(modname) == datas
        cache.add(modname, binaries, datas)
        # This should be extended. Therefore it should be binaries*2
        assert cache.binaries(modname) != binaries
        assert cache.binaries(modname) == binaries * 2



#-----------------------------------------------------------------------------
# Copyright (c) 2020-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
#
# A test script for validation of pkg_resources provider implementation.
#
# The test package has the following structure:
#
# pyi_pkgres_testpkg/
# ├── a.py
# ├── b.py
# ├── __init__.py
# ├── subpkg1
# │   ├── c.py
# │   ├── data
# │   │   ├── entry1.txt
# │   │   ├── entry2.md
# │   │   ├── entry3.rst
# │   │   └── extra
# │   │       └── extra_entry1.json
# │   ├── d.py
# │   └── __init__.py
# ├── subpkg2
# │   ├── __init__.py
# │   ├── mod.py
# │   └── subsubpkg21
# │       ├── __init__.py
# │       └── mod.py
# └── subpkg3
#     ├── _datafile.json
#     └── __init__.py
#
# When run as unfrozen script, this script can be used to check the behavior of "native" providers that come with
# pkg_resources, such as DefaultProvider (for regular packages) and ZipProvider (for eggs).
#
# When run as a frozen application, this script validates the behavior of the frozen provider implemented by
# PyInstaller. Due to transitivity of test results, this script running without errors both as a native script and
# as a frozen application serves as proof of conformance for the PyInstaller's provider.
#
# Wherever the behavior between the native providers is inconsistent, we allow the same leeway for the PyInstaller's
# frozen provider.

import sys
from pkg_resources import resource_exists, resource_isdir, resource_listdir
from pkg_resources import get_provider, DefaultProvider, ZipProvider

pkgname = 'pyi_pkgres_testpkg'

# Identify provider type
provider = get_provider(pkgname)
is_default = isinstance(provider, DefaultProvider)
is_zip = isinstance(provider, ZipProvider)
is_frozen = getattr(sys, 'frozen', False)

assert is_default or is_zip or is_frozen, "Unsupported provider type!"

########################################################################
#                Validate behavior of resource_exists()                #
########################################################################
# Package's directory
#  * DefaultProvider returns True
#  * ZipProvider returns False
#  > PyiFrozenProvider returns True
ret = resource_exists(pkgname, '.')
assert (is_default and ret) or \
       (is_zip and not ret) or \
       (is_frozen and ret)

# Package's directory, with empty path
assert resource_exists(pkgname, '')

# Subpackage's directory (relative to main package):
assert resource_exists(pkgname, 'subpkg1')
assert resource_exists(pkgname, 'subpkg2')
assert resource_exists(pkgname, 'subpkg2/subsubpkg21')
assert resource_exists(pkgname, 'subpkg3')

# Subpackage's directory (relative to subpackage itself):
#  * DefaultProvider returns True
#  * ZipProvider returns False
#  > PyiFrozenProvider returns True
ret = resource_exists(pkgname + '.subpkg1', '.')
assert (is_default and ret) or \
       (is_zip and not ret) or \
       (is_frozen and ret)

# Subpackage's directory (relative to subpackage itself), with empty path:
assert resource_exists(pkgname + '.subpkg1', '')

# Data directory in subpackage
assert resource_exists(pkgname, 'subpkg1/data')
assert resource_exists(pkgname + '.subpkg1', 'data')

# Subdirectory in data directory
assert resource_exists(pkgname, 'subpkg1/data/extra')
assert resource_exists(pkgname + '.subpkg1', 'data/extra')

# File in data directory
assert resource_exists(pkgname, 'subpkg1/data/entry1.txt')

# Deeply nested data file
assert resource_exists(pkgname, 'subpkg1/data/extra/extra_entry1.json')

# A non-existent file/directory - should return False
assert not resource_exists(pkgname, 'subpkg1/non-existant')

# A source script file in package
#  > PyiFrozenProvider returns False because frozen application does not contain source files
ret = resource_exists(pkgname, '__init__.py')
assert (not is_frozen and ret) or \
       (is_frozen and not ret)

# Parent of pacakge's top-level directory
#  * DefaultProvider returns True
#  * ZipProvider returns False
#  > PyiFrozenProvider disallows jumping to parent, and returns False
# NOTE: using .. in path is deprecated (since setuptools 40.8.0) and will raise exception in a future release.
ret = resource_exists(pkgname, '..')
assert (is_default and ret) or \
       (is_zip and not ret) or \
       (is_frozen and not ret)

# Parent of subpackage's directory
#  * DefaultProvider returns True
#  * ZipProvider returns False
#  > PyiFrozenProvider disallows jumping to parent, and returns False
# NOTE: using .. in path is deprecated (since setuptools 40.8.0) and will raise exception in a future release.
ret = resource_exists(pkgname + '.subpkg1', '..')
assert (is_default and ret) or \
       (is_zip and not ret) or \
       (is_frozen and not ret)

# Submodule in main package
ret = resource_exists(pkgname + '.a', '.')
assert (is_default and ret) or \
       (is_zip and not ret) or \
       (is_frozen and ret)

# Submodule in main package, with empty path
assert resource_exists(pkgname + '.a', '')

# Submodule in subpackage
ret = resource_exists(pkgname + '.subpkg1.c', '.')
assert (is_default and ret) or \
       (is_zip and not ret) or \
       (is_frozen and ret)

# Submodule in subpackage, with empty path
assert resource_exists(pkgname + '.subpkg1.c', '')

########################################################################
#                Validate behavior of resource_isdir()                 #
########################################################################
# Package's directory
#  * DefaultProvider returns True
#  * ZipProvider returns False
#  > PyiFrozenProvider returns True
ret = resource_isdir(pkgname, '.')
assert (is_default and ret) or \
       (is_zip and not ret) or \
       (is_frozen and ret)

# Package's directory, with empty path
assert resource_isdir(pkgname, '')

# Subpackage's directory (relative to main package):
#  * both DefaultProvider and ZipProvider return True
assert resource_isdir(pkgname, 'subpkg1')
assert resource_isdir(pkgname, 'subpkg2')
assert resource_isdir(pkgname, 'subpkg2/subsubpkg21')
assert resource_isdir(pkgname, 'subpkg3')

# Subpackage's directory (relative to subpackage itself):
#  * DefaultProvider returns True
#  * ZipProvider returns False
#  > PyiFrozenProvider returns True
ret = resource_isdir(pkgname + '.subpkg1', '.')
assert (is_default and ret) or \
       (is_zip and not ret) or \
       (is_frozen and ret)

# Subpackage's directory (relative to subpackage itself), with empty path:
assert resource_isdir(pkgname + '.subpkg1', '')

# Data directory in subpackage
assert resource_isdir(pkgname, 'subpkg1/data')
assert resource_isdir(pkgname + '.subpkg1', 'data')

# Subdirectory in data directory
assert resource_isdir(pkgname, 'subpkg1/data/extra')
assert resource_isdir(pkgname + '.subpkg1', 'data/extra')

# File in data directory - should return False
assert not resource_isdir(pkgname, 'subpkg1/data/entry1.txt')

# Deeply nested data file - should return False
assert not resource_isdir(pkgname, 'subpkg1/data/extra/extra_entry1.json')

# A non-existent file-directory - should return False
assert not resource_isdir(pkgname, 'subpkg1/non-existant')

# A source script file in package - should return False
# NOTE: PyFrozenProvider returns False because the file does not exist.
assert not resource_isdir(pkgname, '__init__.py')

# Parent of package's top-level directory
#  * DefaultProvider returns True
#  * ZipProvider returns False
#  > PyiFrozenProvider disallows jumping to parent, and returns False
# NOTE: using .. in path is deprecated (since setuptools 40.8.0) and will raise exception in a future release.
ret = resource_isdir(pkgname, '..')
assert (is_default and ret) or \
       (is_zip and not ret) or \
       (is_frozen and not ret)

# Parent of subpacakge's directory
#  * DefaultProvider returns True
#  * ZipProvider returns False
#  > PyiFrozenProvider disallows jumping to parent, and returns False
# NOTE: using .. in path is deprecated (since setuptools 40.8.0) and will raise exception in a future release.
ret = resource_isdir(pkgname + '.subpkg1', '..')
assert (is_default and ret) or \
       (is_zip and not ret) or \
       (is_frozen and not ret)

# Submodule in main package
ret = resource_isdir(pkgname + '.a', '.')
assert (is_default and ret) or \
       (is_zip and not ret) or \
       (is_frozen and ret)

# Submodule in main package, with empty path
assert resource_isdir(pkgname + '.a', '')

# Submodule in subpackage
ret = resource_isdir(pkgname + '.subpkg1.c', '.')
assert (is_default and ret) or \
       (is_zip and not ret) or \
       (is_frozen and ret)

# Submodule in subpackage, with empty path
assert resource_isdir(pkgname + '.subpkg1.c', '')


########################################################################
#               Validate behavior of resource_listdir()                #
########################################################################
# A helper for resource_listdir() tests.
def _listdir_test(pkgname, path, expected):
    # For frozen application, remove .py files from expected results
    if is_frozen:
        expected = [x for x in expected if not x.endswith('.py')]
    # List the content
    content = resource_listdir(pkgname, path)
    # Ignore pycache
    if '__pycache__' in content:
        content.remove('__pycache__')
    assert sorted(content) == sorted(expected)


# List package's top-level directory
#  * DefaultProvider lists the directory
#  * ZipProvider returns empty list
#  > PyiFrozenProvider lists the directory, but does not provide source .py files
if is_zip:
    expected = []
else:
    expected = ['__init__.py', 'a.py', 'b.py', 'subpkg1', 'subpkg2', 'subpkg3']
_listdir_test(pkgname, '.', expected)

# List package's top-level directory, with empty path
#  > PyiFrozenProvider lists the directory, but does not provide source .py files
expected = ['__init__.py', 'a.py', 'b.py', 'subpkg1', 'subpkg2', 'subpkg3']
_listdir_test(pkgname, '', expected)

# List subpackage's directory (relative to main package)
#  > PyiFrozenProvider lists the directory, but does not provide source .py files
expected = ['__init__.py', 'c.py', 'd.py', 'data']
_listdir_test(pkgname, 'subpkg1', expected)

# List data directory in subpackage (relative to main package)
expected = ['entry1.txt', 'entry2.md', 'entry3.rst', 'extra']
_listdir_test(pkgname, 'subpkg1/data', expected)

# List data directory in subpackage (relative to subpackage itself)
expected = ['entry1.txt', 'entry2.md', 'entry3.rst', 'extra']
_listdir_test(pkgname + '.subpkg1', 'data', expected)

# List data in subdirectory of data directory in subpackage
expected = ['extra_entry1.json']
_listdir_test(pkgname + '.subpkg1', 'data/extra', expected)

# Attempt to list a file (existing resource but not a directory).
#  * DefaultProvider raises NotADirectoryError
#  * ZipProvider returns empty list
#  > PyiFrozenProvider returns empty list
try:
    content = resource_listdir(pkgname + '.subpkg1', 'data/entry1.txt')
except NotADirectoryError:
    assert is_default
except Exception:
    raise
else:
    assert (is_zip or is_frozen) and content == []

# Attempt to list an non-existent directory in main package.
#  * DefaultProvider raises FileNotFoundError
#  * ZipProvider returns empty list
#  > PyiFrozenProvider returns empty list
try:
    content = resource_listdir(pkgname, 'non-existant')
except FileNotFoundError:
    assert is_default
except Exception:
    raise
else:
    assert (is_zip or is_frozen) and content == []

# Attempt to list an non-existent directory in subpackage
#  * DefaultProvider raises FileNotFoundError
#  * ZipProvider returns empty list
#  > PyiFrozenProvider returns empty list
try:
    content = resource_listdir(pkgname + '.subpkg1', 'data/non-existant')
except FileNotFoundError:
    assert is_default
except Exception:
    raise
else:
    assert (is_zip or is_frozen) and content == []

# Attempt to list pacakge's parent directory
#  * DefaultProvider actually lists the parent directory
#  * ZipProvider returns empty list
#  > PyiFrozenProvider disallows jumping to parent, and returns empty list
# NOTE: using .. in path is deprecated (since setuptools 40.8.0) and
# will raise exception in a future release
content = resource_listdir(pkgname, '..')
assert (is_default and pkgname in content) or \
       (is_zip and content == []) or \
       (is_frozen and content == [])

# Attempt to list subpackage's parent directory
#  * DefaultProvider actually lists the parent directory
#  * ZipProvider returns empty list
#  > PyiFrozenProvider disallows jumping to parent, and returns False
# NOTE: using .. in path is deprecated (since setuptools 40.8.0) and will raise exception in a future release.
if is_default:
    expected = ['__init__.py', 'a.py', 'b.py', 'subpkg1', 'subpkg2', 'subpkg3']
else:
    expected = []
_listdir_test(pkgname + '.subpkg1', '..', expected)

# Attempt to list directory of subpackage that has no data files or directories (relative to main package)
expected = ['__init__.py', 'mod.py', 'subsubpkg21']
_listdir_test(pkgname, 'subpkg2', expected)

# Attempt to list directory of subpackage that has no data files or directories (relative to subpackage itself)
expected = ['__init__.py', 'mod.py', 'subsubpkg21']
_listdir_test(pkgname + '.subpkg2', '', expected)  # empty path!

# Attempt to list directory of subsubpackage that has no data files/directories (relative to main package)
expected = ['__init__.py', 'mod.py']
_listdir_test(pkgname, 'subpkg2/subsubpkg21', expected)

# Attempt to list directory of subsubpackage that has no data files/directories (relative to parent subpackage)
expected = ['__init__.py', 'mod.py']
_listdir_test(pkgname + '.subpkg2', 'subsubpkg21', expected)

# Attempt to list directory of subsubpackage that has no data files/directories (relative to subsubpackage itself)
expected = ['__init__.py', 'mod.py']
_listdir_test(pkgname + '.subpkg2.subsubpkg21', '', expected)  # empty path!

# Attempt to list submodule in main package - should give the same results as listing the package itself
assert sorted(resource_listdir(pkgname + '.a', '')) == \
       sorted(resource_listdir(pkgname, ''))  # empty path!

# Attempt to list submodule in subpackage - should give the same results as listing the subpackage itself
assert sorted(resource_listdir(pkgname + '.subpkg1.c', '')) == \
       sorted(resource_listdir(pkgname + '.subpkg1', ''))  # empty path!



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Functional tests for SciPy.
"""

import sys

import pytest

from PyInstaller.utils.tests import importorskip, xfail, onedir_only

pytestmark = [
    importorskip('scipy'),
    xfail(
        sys.version_info[:3] == (3, 12, 0),
        reason='SciPy is broken with PyInstaller and python 3.12.0 (#7992).',
    ),
]


# Basic import test for each scipy module, to ensure that each module is importable on its own. Due to amount of tests,
# run them only in onedir mode.
@pytest.mark.parametrize(
    'module', [
        'scipy',
        'scipy.cluster',
        'scipy.constants',
        'scipy.datasets',
        'scipy.fft',
        'scipy.fftpack',
        'scipy.integrate',
        'scipy.interpolate',
        'scipy.io',
        'scipy.linalg',
        'scipy.misc',
        'scipy.ndimage',
        'scipy.odr',
        'scipy.optimize',
        'scipy.signal',
        'scipy.sparse',
        'scipy.spatial',
        'scipy.special',
        'scipy.stats',
    ]
)
@onedir_only
def test_scipy(pyi_builder, module):
    pyi_builder.test_source(f"""
        import {module}
        """)



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2024, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader. Additional
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# --- Copyright Disclaimer ---
#
# An earlier copy of this hook has been submitted to the NumPy project, where it was integrated in v1.23.0rc1
# (https://github.com/numpy/numpy/pull/20745), under terms and conditions outlined in their repository [1].
#
# A special provision is hereby granted to the NumPy project that allows the NumPy copy of the hook to incorporate the
# changes made to this (PyInstaller's) copy of the hook, subject to their licensing terms as opposed to PyInstaller's
# (stricter) licensing terms.
#
# .. refs:
#
#   [1] NumPy's license: https://github.com/numpy/numpy/blob/master/LICENSE.txt

# NOTE: when comparing the contents of this hook and the NumPy version of the hook (for example, to port changes), keep
# in mind that this copy is PyInstaller-centric - it caters to the version of PyInstaller it is bundled with, but needs
# to account for different behavior of different NumPy versions. In contrast, the NumPy copy of the hook caters to the
# version of NumPy it is bundled with, but should account for behavior differences in different PyInstaller versions.

# Override the default hook priority so that our copy of hook is used instead of NumPy's one (which has priority 0,
# the default for upstream hooks).
# $PyInstaller-Hook-Priority: 1

from PyInstaller import compat
from PyInstaller.utils.hooks import (
    get_installer,
    collect_dynamic_libs,
)

from packaging.version import Version

numpy_version = Version(compat.importlib_metadata.version("numpy")).release
numpy_installer = get_installer('numpy')

hiddenimports = []
datas = []
binaries = []

# Collect shared libraries that are bundled inside the numpy's package directory. With PyInstaller 6.x, the directory
# layout of collected shared libraries should be preserved (to match behavior of the binary dependency analysis). In
# earlier versions of PyInstaller, it was necessary to collect the shared libraries into application's top-level
# directory (because that was also what binary dependency analysis in PyInstaller < 6.0 did).
binaries += collect_dynamic_libs("numpy")

# Check if we are using Anaconda-packaged numpy
if numpy_installer == 'conda':
    # Collect DLLs for NumPy and its dependencies (MKL, OpenBlas, OpenMP, etc.) from the communal Conda bin directory.
    from PyInstaller.utils.hooks import conda_support
    datas += conda_support.collect_dynamic_libs("numpy", dependencies=True)

# NumPy 1.26 started using `delvewheel` for its Windows PyPI wheels. While contemporary PyInstaller versions
# automatically pick up DLLs from external `numpy.libs` directory, this does not work on Anaconda python 3.8 and 3.9
# due to defunct `os.add_dll_directory`, which forces `delvewheel` to use the old load-order file approach. So we need
# to explicitly ensure that load-order file as well as DLLs are collected.
if compat.is_win and numpy_version >= (1, 26) and numpy_installer == 'pip':
    from PyInstaller.utils.hooks import collect_delvewheel_libs_directory
    datas, binaries = collect_delvewheel_libs_directory("numpy", datas=datas, binaries=binaries)

# Submodules PyInstaller cannot detect (probably because they are only imported by extension modules, which PyInstaller
# cannot read).
if numpy_version >= (2, 0):
    # In v2.0.0, `numpy.core` was renamed to `numpy._core`.
    # See https://github.com/numpy/numpy/commit/47b70cbffd672849a5d3b9b6fa6e515700460fd0
    hiddenimports += ['numpy._core._dtype_ctypes', 'numpy._core._multiarray_tests']
else:
    hiddenimports += ['numpy.core._dtype_ctypes']

    # See https://github.com/numpy/numpy/commit/99104bd2d0557078d7ea9a590129c87dd63df623
    if numpy_version >= (1, 25):
        hiddenimports += ['numpy.core._multiarray_tests']

# Starting with v2.3.0, we need to add `numpy._core._exceptions` to hiddenimports; in previous versions, this module
# was picked up due to explicit import in `numpy._core._methods`, which was removed as part of cleanup in
# https://github.com/numpy/numpy/commit/a51a4f5c10aa9b7962ff1e7e9b5f9b7d91c51489
if numpy_version >= (2, 3, 0):
    hiddenimports += ['numpy._core._exceptions']

# This hidden import was removed from NumPy hook in v1.25.0 (https://github.com/numpy/numpy/pull/22666). According to
# comment in the linked PR, it should have been unnecessary since v1.19.
if compat.is_conda and numpy_version < (1, 19):
    hiddenimports += ["six"]

# Remove testing and building code and packages that are referenced throughout NumPy but are not really dependencies.
excludedimports = [
    "scipy",
    "pytest",
    "nose",
    "f2py",
    "setuptools",
]

# As of v1.22.0, numpy.testing (imported for example by some scipy modules) requires numpy.distutils and distutils.
# This was due to numpy.testing adding import of numpy.testing._private.extbuild, which in turn imported numpy.distutils
# and distutils. These imports were moved into functions that require them in v1.22.2 and v.1.23.0.
# See: https://github.com/numpy/numpy/pull/20831 and https://github.com/numpy/numpy/pull/20906
# So we can exclude them for all numpy versions except for v1.22.0 and v1.22.1 - the main motivation is to avoid pulling
# in `setuptools` (which nowadays provides its vendored version of `distutils`).
if numpy_version < (1, 22, 0) or numpy_version > (1, 22, 1):
    excludedimports += [
        "distutils",
        "numpy.distutils",
    ]

# In numpy v2.0.0, numpy.f2py submodule has been added to numpy's `__all__` attribute. Therefore, using
# `from numpy import *` leads to an error if `numpy.f2py` is excluded (seen in scipy 1.14). The exclusion in earlier
# releases was not reported to cause any issues, so keep it around. Although it should be noted that it does break an
# explicit import (i.e., `import numpy.f2py`) from user's code as well, because it prevents collection of other
# submodules from `numpy.f2py`.
if numpy_version < (2, 0):
    excludedimports += [
        "numpy.f2py",
    ]



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# PyInstaller always loads modules from the embedded archive before looking at sys.path.
#
# This tests creates a module with the same name as the one in the embedded archive. The frozen application
# should always load the module from the embedded archive.

import os
import sys
import importlib

# Create module.
txt = """
x = %d
"""
mod_filename = os.path.join(sys._MEIPASS, 'data_reload.py')
with open(mod_filename, 'w') as f:
    f.write(txt % 2)

# Import created module.
import data_reload  # noqa: E402

orig_x = data_reload.x
print(('data_reload.x is %s' % data_reload.x))

# Modify code of module - increment x.
with open(mod_filename, 'w') as f:
    f.write(txt % (data_reload.x + 1))

# Reload module.
importlib.reload(data_reload)
print(('data_reload.x is now %s' % data_reload.x))
# The value of 'x' should be the orig_x + 1.
assert data_reload.x == orig_x + 1



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
When frozen, a module that dynamically recreates itself at runtime (by replacing itself in sys.modules) should be
returned by __import__ statement.

This example should return True:
    >>> sys.modules[<dynamic_module>] is __import__(<dynamic_module>)
    True
"""

import sys
import types

# The DynamicModule should override this attribute.
foo = None


class DynamicModule(types.ModuleType):
    __file__ = __file__

    def __init__(self, name):
        super().__init__(name)
        self.foo = "A new value!"


# Replace module 'pyi_testmod_dynamic' by class DynamicModule.
sys.modules[__name__] = DynamicModule(__name__)



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import codecs
import locale
import os
import sys
import pathlib
import subprocess
import re

import pytest

from PyInstaller import isolated
from PyInstaller.compat import is_cygwin, is_darwin, is_termux, is_win
from PyInstaller.utils.tests import importorskip, importable, skipif, xfail, onedir_only, onefile_only


def test_run_from_path_environ(pyi_builder):
    pyi_builder.test_script('pyi_absolute_python_path.py', run_from_path=True)


@pytest.mark.linux
def test_absolute_ld_library_path(pyi_builder):
    pyi_builder.test_script('pyi_absolute_ld_library_path.py')


def test_absolute_python_path(pyi_builder):
    pyi_builder.test_script('pyi_absolute_python_path.py')


# Test that on linux, the bootloader preserves the original process name when launching a child process (onefile mode)
# or when restarting itself (onedir mode). This is relevant when executable is launched via symbolic link that has
# different basename than the target executable; the process name (as stored in `/proc/self/status` or retrieved via
# `prtctl()` with `PR_GET_NAME`) should be the basename of the symbolic link!
#
# Test both variant without splash screen (the default) and with splash screen enabled; in the latter case, the onefile
# parent process needs to restart itself, so we need to also test process name preservation in that codepath.
@pytest.mark.linux
@skipif(not os.path.exists('/proc/self/status'), reason='/proc/self/status does not exist')
@pytest.mark.parametrize('enable_splash', [False, True], ids=['nosplash', 'splash'])
def test_symlink_basename_is_kept(pyi_builder, tmp_path, enable_splash):
    if enable_splash:
        if not importable("tkinter"):
            pytest.skip("Needs tkinter")
        splash_image = pathlib.Path(__file__).parent / 'data' / 'splash' / 'image.png'
        extra_args = ['--splash', str(splash_image)]
    else:
        extra_args = []

    # Build the test program and run it with actual executable for sanity check.
    pyi_builder.test_source(
        """
        import os
        import sys

        # Read the process name from `/proc/self/status`.
        with open('/proc/self/status', 'r') as fh:
            for line in fh.readlines():
                if line.startswith('Name:'):
                    name_from_proc = line.split(":")[1].strip()
                    break

        # The name passed via argv.
        name_from_argv = os.path.basename(sys.argv[0])

        # The process name /proc/self/status is truncated to 15 characters, so truncated argv-based name for comparison.
        truncated_name_from_argv = name_from_argv[:15]

        print(f"ARGV: {name_from_argv!r} (complete)")
        print(f"ARGV: {truncated_name_from_argv!r} (truncated")
        print(f"PROC: {name_from_proc!r}")

        assert truncated_name_from_argv == name_from_proc
        """,
        pyi_args=extra_args,
    )

    # Find executable
    exes = pyi_builder._find_executables('test_source')
    assert len(exes) == 1
    program_exe = exes[0]

    # Create and test with various symbolic links (short name, long name, sub-directory).
    SYMLINK_NAMES = ["symlink", "very_long_name_in_symlink", "sub/dir/program"]
    for symlink_name in SYMLINK_NAMES:
        # Print banner to both stdout and stderr, as they are captured separately.
        print(f"=== running test with {symlink_name!r} ===", file=sys.stderr)
        print(f"=== running test with {symlink_name!r} ===", file=sys.stdout)

        # Crate symbolic link
        symlinked_exe = tmp_path / symlink_name
        symlinked_exe.parent.mkdir(parents=True, exist_ok=True)

        symlinked_exe.symlink_to(program_exe)

        # Run
        subprocess.run([symlinked_exe], check=True)


@onedir_only
def test_pyz_as_external_file(pyi_builder, monkeypatch):
    # This tests the not well documented and seldom used feature of having the PYZ-archive in a separate file (.pkg).

    def MyEXE(*args, **kwargs):
        kwargs['append_pkg'] = False
        return EXE(*args, **kwargs)

    import PyInstaller.building.build_main
    EXE = PyInstaller.building.build_main.EXE
    monkeypatch.setattr('PyInstaller.building.build_main.EXE', MyEXE)

    pyi_builder.test_source("print('Hello Python!')")


def test_celementtree(pyi_builder):
    pyi_builder.test_source("""
        from xml.etree.cElementTree import ElementTree
        print('OK')
        """)


# Test a build with some complexity with the ``noarchive`` debug option.
def test_noarchive(pyi_builder):
    pyi_builder.test_source("from xml.etree.cElementTree import ElementTree", pyi_args=['--debug=noarchive'])


@importorskip('codecs')
def test_codecs(pyi_builder):
    pyi_builder.test_script('pyi_codecs.py')


def test_compiled_filenames(pyi_builder):
    pyi_builder.test_source(
        """
        import pyi_dummy_module
        from os.path import isabs

        assert not isabs(pyi_dummy_module.dummy.__code__.co_filename), (
            "pyi_dummy_module.dummy.__code__.co_filename has compiled filename: %s" %
            (pyi_dummy_module.dummy.__code__.co_filename, )
        )
        assert not isabs(pyi_dummy_module.DummyClass.dummyMethod.__code__.co_filename), (
            "pyi_dummy_module.DummyClass.dummyMethod.__code__.co_filename has compiled filename: %s" %
            (pyi_dummy_module.DummyClass.dummyMethod.__code__.co_filename, )
        )
        """
    )


def test_decoders_ascii(pyi_builder):
    pyi_builder.test_source(
        """
        # Convert type 'bytes' to type 'str'.
        assert b'foo'.decode('ascii') == 'foo'
        """
    )


def test_distutils_submod(pyi_builder):
    # Test import of submodules of distutils package.
    # PyI fails to include `distutils.version` when running from virtualenv.
    pyi_builder.test_source("""
        from distutils.version import LooseVersion
        """)


def test_dynamic_module(pyi_builder):
    pyi_builder.test_source(
        """
        import pyi_testmod_dynamic

        # The value 'foo' should  not be None.
        print("'foo' value: %s" % pyi_testmod_dynamic.foo)
        assert pyi_testmod_dynamic.foo is not None
        assert pyi_testmod_dynamic.foo == 'A new value!'
        """
    )


def test_email(pyi_builder):
    pyi_builder.test_source(
        """
        from email import utils
        from email.header import Header
        from email.mime.text import MIMEText
        from email.mime.multipart import MIMEMultipart
        from email.mime.nonmultipart import MIMENonMultipart
        """
    )


def test_filename(pyi_builder):
    pyi_builder.test_script('pyi_filename.py')


def test_getfilesystemencoding(pyi_builder):
    pyi_builder.test_script('pyi_getfilesystemencoding.py')


def test_helloworld(pyi_builder):
    pyi_builder.test_source("print('Hello Python!')")


def test_module__file__attribute(pyi_builder):
    pyi_builder.test_script('pyi_module__file__attribute.py')


def test_module_attributes(tmp_path, pyi_builder):
    # Modules to test
    MODULE_NAMES = [
        "xml.etree.ElementTree",  # pure-Python module
        "xml.etree.cElementTree",  # binary extension
    ]

    # Get module attributes in unfrozen python
    @isolated.decorate
    def _module_attributes(module_names):
        import importlib

        output = {}
        for module_name in module_names:
            module = importlib.import_module(module_name)
            output[module_name] = sorted(dir(module))

        return output

    output_unfrozen = _module_attributes(MODULE_NAMES)

    # Get attributes in frozen test
    output_file = tmp_path / "output.json"
    app_args = [str(output_file), *MODULE_NAMES]

    pyi_args = []
    for module_name in MODULE_NAMES:
        pyi_args += ['--hidden-import', module_name]

    pyi_builder.test_source(
        """
        import importlib
        import json
        import sys

        if len(sys.argv) < 3:
            print(f"Usage: {sys.argv[0]} <output_filename> <module_name> [module_name] [...]")
            sys.exit(1)

        output_filename = sys.argv[1]
        module_names = sys.argv[2:]

        output = {}
        for module_name in module_names:
            module = importlib.import_module(module_name)
            output[module_name] = sorted(dir(module))

        with open(output_filename, "w") as fp:
            json.dump(output, fp)
        """,
        pyi_args=pyi_args,
        app_args=app_args,
    )

    with open(output_file, 'r') as fp:
        import json
        output_frozen = json.load(fp)

    for module_name in MODULE_NAMES:
        print("", file=sys.stderr)
        print(f"Comparing attributes of {module_name!r}:", file=sys.stderr)
        print(f"Unfrozen: {output_unfrozen[module_name]!r}", file=sys.stderr)
        print(f"Frozen: {output_frozen[module_name]}", file=sys.stderr)

        assert output_frozen[module_name] == output_unfrozen[module_name]


def test_module_reload(pyi_builder):
    pyi_builder.test_script('pyi_module_reload.py')


def test_chdir_meipass(pyi_builder):
    # Ensure meipass dir exists.
    pyi_builder.test_source(
        """
        import os, sys
        os.chdir(sys._MEIPASS)
        print(os.getcwd())
        """
    )


def test_option_exclude_module(pyi_builder):
    """
    Test to ensure that when using option --exclude-module=xml.sax
    the module 'xml.sax' won't be bundled.
    """
    pyi_builder.test_source(
        """
        try:
            import xml.sax
            # Option --exclude-module=xml.sax did not work and the module
            # was successfully imported.
            raise SystemExit('Module xml.sax was excluded but it is '
                             'bundled with the executable.')
        except ImportError:
            # The Import error is expected since PyInstaller should
            # not bundle 'xml.sax' module.
            pass
        """,
        pyi_args=['--exclude-module', 'xml.sax']
    )


def test_option_verbose(pyi_builder, monkeypatch):
    """
    Test to ensure that option V can be set and has effect.
    """

    # This option is like 'python -v' - trace import statements.
    # 'None' should be allowed or '' also.

    def MyEXE(*args, **kwargs):
        args = list(args)
        args.append([('v', None, 'OPTION')])
        return EXE(*args, **kwargs)

    import PyInstaller.building.build_main
    EXE = PyInstaller.building.build_main.EXE
    monkeypatch.setattr('PyInstaller.building.build_main.EXE', MyEXE)

    pyi_builder.test_source(
        """
        print('test - PYTHONVERBOSE - trace import statements')
        import re # just import anything
        print('test - done')
        """
    )


def test_option_w_unset(pyi_builder):
    """
    Test to ensure that option W is not set by default.
    """
    pyi_builder.test_source("""
        import sys
        assert 'ignore' not in sys.warnoptions
        """)


def test_option_w_ignore(pyi_builder, monkeypatch, capsys):
    """
    Test to ensure that option W can be set.
    """
    def MyEXE(*args, **kwargs):
        args = list(args)
        args.append([('W ignore', '', 'OPTION')])
        return EXE(*args, **kwargs)

    import PyInstaller.building.build_main
    EXE = PyInstaller.building.build_main.EXE
    monkeypatch.setattr('PyInstaller.building.build_main.EXE', MyEXE)

    pyi_builder.test_source("""
        import sys
        assert 'ignore' in sys.warnoptions
        """)

    _, err = capsys.readouterr()
    assert "'import warnings' failed" not in err


@pytest.mark.parametrize("distutils", [False, True], ids=["sysconfig", "distutils.sysconfig"])
def test_python_makefile(pyi_builder, distutils):
    """
    Tests hooks for ``sysconfig`` and its near-duplicate ``distutils.sysconfig``. Raises an import error if we fail
    to collect the special module that contains the details from pyconfig.h and the Makefile.
    """
    # Ideally we would test that the contents of `sysconfig.get_config_vars()` dict are the same frozen vs. unfrozen,
    # but because some values are paths into a Python installation's guts, these will point into the frozen app when
    # frozen, and therefore not match. Without some fiddly filtering of the paths, this is impossible.

    # As a compromise, test that the dictionary keys are the same to be sure that there is no conditional initialisation
    # of get_config_vars(). I.e., that get_config_vars() does not silently return an empty dictionary if it cannot find
    # the information it needs.
    if distutils:
        from distutils import sysconfig
        import_preamble = 'from distutils '
    else:
        import sysconfig
        import_preamble = ''
    unfrozen_keys = sorted(sysconfig.get_config_vars().keys())

    pyi_builder.test_source(
        f"""
        # The error is raised immediately on import.
        {import_preamble}import sysconfig

        # But just in case, Python later opt for some lazy loading, force
        # configuration retrieval:
        from pprint import pprint
        pprint(sysconfig.get_config_vars())

        unfrozen_keys = {unfrozen_keys}
        assert sorted(sysconfig.get_config_vars()) == unfrozen_keys
        """
    )


def test_set_icon(pyi_builder, data_dir):
    if is_win:
        args = ['--icon', str(data_dir / 'pyi_icon.ico')]
    elif is_darwin:
        # On macOS icon is applied only for windowed mode.
        args = ['--windowed', '--icon', str(data_dir / 'pyi_icon.icns')]
    else:
        pytest.skip('option --icon works only on Windows and macOS')
    os.chdir(os.path.expanduser("~"))
    pyi_builder.test_source("print('Hello Python!')", pyi_args=args)


def test_python_home(pyi_builder):
    pyi_builder.test_script('pyi_python_home.py')


@pytest.mark.parametrize('stream', ['stdout', 'stderr'])
def test_standard_stream_encoding(stream, tmp_path, pyi_builder):
    # NOTE: '-s' option to pytest disables output capturing, changing this test's result:
    # without -s: pytest process changes its own stdout encoding to 'UTF-8' to capture output. subprocess spawned by
    #             pytest has stdout encoding 'cp1252', which is an ANSI codepage. test fails as they do not match.
    # with -s:    pytest process has stdout encoding from windows terminal, which is an OEM codepage. spawned
    #             subprocess has the same encoding. test passes.

    # A test program that dumps encoding of the stream into specified file.
    frozen_encoding_file = tmp_path / 'frozen_encoding.txt'
    pyi_builder.test_source(
        f"""
        import sys

        encoding = str(sys.{stream}.encoding)
        if len(sys.argv) >= 2:
            with open(sys.argv[1], 'w', encoding='utf-8') as fp:
                fp.write(encoding)
        else:
            print(encoding)
        """,
        app_args=[str(frozen_encoding_file)]
    )
    frozen_encoding = frozen_encoding_file.read_text(encoding='utf-8')
    print(f"Frozen encoding: {frozen_encoding}")

    # For non-interactive stdout/stderr, assume locale encoding (on Windows, this will be ANSI codepage). This fixes the
    # test when running with pytest and capturing output.
    unfrozen_stream = getattr(sys, stream)
    unfrozen_encoding = (
        str(unfrozen_stream.encoding) if unfrozen_stream.isatty() else locale.getpreferredencoding(False)
    )
    print(f"Unfrozen encoding: {frozen_encoding}")

    # Normalize encoding names - "UTF-8" should be the same as "utf8".
    unfrozen_encoding = codecs.lookup(unfrozen_encoding).name
    frozen_encoding = codecs.lookup(frozen_encoding).name

    assert frozen_encoding == unfrozen_encoding


def test_site_module_disabled(pyi_builder):
    pyi_builder.test_script('pyi_site_module_disabled.py')


def test_time_module(pyi_builder):
    pyi_builder.test_source("""
        import time
        print(time.strptime(time.ctime()))
        """)


@pytest.mark.darwin
@pytest.mark.linux
def test_time_module_localized(pyi_builder, monkeypatch):
    # This checks that functions 'time.ctime()' and 'time.strptime()' use the same locale. There was an issue with
    # bootloader where every function was using different locale:
    # time.ctime was using 'C'
    # time.strptime was using 'xx_YY' from the environment.
    monkeypatch.setenv('LC_ALL', 'cs_CZ.UTF-8')
    pyi_builder.test_source("""
        import time
        print(time.strptime(time.ctime()))
        """)


# Check that `locale.getlocale()` in frozen application returns user-preferred locale (i.e., that user-preferred
# locale is set in the bootloader during python interpreter setup). The test ensures that the user-preferred locale
# is  set *at all* (see #8305), as well as that it matches the environment variables.
@pytest.mark.darwin
@pytest.mark.linux
def test_user_preferred_locale(pyi_builder):
    # NOTE: this runs the program without arguments, which checks that locale is set at all.
    pyi_builder.test_source(
        """
        import sys
        import locale

        user_locale = locale.getlocale()
        print(f"User locale: {user_locale}", file=sys.stderr)

        if len(sys.argv) == 1:
            # No arguments - check that locale is set at all; the tuple must have two elements, and neither of them is
            # None.
            if (len(user_locale) != 2) or (user_locale[0] is None) or (user_locale[1] is None):
                raise Exception(f"Invalid user locale: {user_locale!r}")
        elif len(sys.argv) == 2:
            expected_locale = tuple(sys.argv[1].split('.'))
            print(f"Expected locale: {expected_locale}", file=sys.stderr)
            if user_locale != expected_locale:
                raise Exception(f"Unexpected user locale: {user_locale!r} (expected: {expected_locale!r})")
        else:
            print(f"Usage: {sys.argv[0]} [expected_locale]")
            sys.exit(1)

        print("OK!", file=sys.stderr)
        """
    )

    # Find executable and run additional tests with locale set via LC_ALL.
    if is_termux:
        return

    exes = pyi_builder._find_executables('test_source')
    assert len(exes) == 1

    test_locales = [
        "en_US.UTF-8",
        "en_US.ISO8859-1",
        "sl_SI.UTF-8",
        "sl_SI.ISO8859-2",
    ]

    for test_locale in test_locales:
        print(f"Running test with locale: {test_locale!r}...", file=sys.stderr)
        env = {
            "LC_ALL": test_locale,
        }
        subprocess.run([exes[0], test_locale], check=True, env=env)


def test_xmldom_module(pyi_builder):
    pyi_builder.test_source(
        """
        print('Importing xml.dom')
        from xml.dom import pulldom
        print('Importing done')
        """
    )


def test_threading_module(pyi_builder):
    pyi_builder.test_source(
        """
        import threading
        import sys

        print('See stderr for messages')
        def print_(*args): print(*args, file=sys.stderr)

        def doit(nm):
            print_(nm, 'started')
            import pyi_testmod_threading
            try:
                print_(nm, pyi_testmod_threading.x)
            finally:
                print_(nm, pyi_testmod_threading)

        t1 = threading.Thread(target=doit, args=('t1',))
        t2 = threading.Thread(target=doit, args=('t2',))
        t1.start()
        t2.start()
        doit('main')
        t1.join() ; print_('t1 joined')
        t2.join() ; print_('t2 joined')
        print_('finished.')
        """
    )


def test_threading_module2(pyi_builder):
    pyi_builder.test_script('pyi_threading_module2.py')


def test_argument(pyi_builder):
    pyi_builder.test_source(
        """
        import sys
        assert sys.argv[1] == "--argument", "sys.argv[1] was %r, expected %r" % (sys.argv[1], "--argument")
        """,
        app_args=["--argument"]
    )


@pytest.mark.win32
def test_renamed_exe(pyi_builder):
    _old_find_executables = pyi_builder._find_executables

    def _find_executables(name):
        old_executables = _old_find_executables(name)
        new_executables = []
        for old_exe in old_executables:
            old_exe_path = pathlib.Path(old_exe)
            new_exe_path = old_exe_path.with_name(f"renamed_{old_exe_path.name}")
            old_exe_path.rename(new_exe_path)

            new_executables.append(str(new_exe_path))
        return new_executables

    pyi_builder._find_executables = _find_executables
    pyi_builder.test_source("print('Hello Python!')")


def test_spec_with_utf8(pyi_builder_spec):
    pyi_builder_spec.test_spec('spec-with-utf8.spec')


@pytest.mark.darwin
def test_osx_override_info_plist(pyi_builder_spec):
    pyi_builder_spec.test_spec('pyi_osx_override_info_plist.spec')


def test_hook_collect_submodules(pyi_builder, script_dir):
    # This is designed to test the operation of PyInstaller.utils.hook.collect_submodules. To do so:
    #
    # 1. It imports the dummy module pyi_collect_submodules_mod, which contains nothing.
    # 2. This causes hook-pyi_collect_submodules_mod.py to be run, which collects some dummy submodules. In this case,
    #    it collects from modules/pyi_testmod_relimp.
    # 3. Therefore, we should be able to find hidden imports under pyi_testmod_relimp.
    pyi_builder.test_source(
        """
        import pyi_collect_submodules_mod
        __import__('pyi_testmod_relimp.B.C')
        """,
        pyi_args=['--additional-hooks-dir', str(script_dir / 'pyi_hooks')]
    )


# Test that PyInstaller can handle a script with an arbitrary extension/suffix.
def test_arbitrary_entry_point_script_suffix(pyi_builder):
    pyi_builder.test_script('pyi_arbitrary_ext.foo')


# Test that PyInstaller can handle a script with no extension/suffix.
def test_no_entry_point_script_suffix(pyi_builder):
    pyi_builder.test_script('pyi_no_ext')


@onefile_only
def test_option_runtime_tmpdir(pyi_builder):
    """
    Test to ensure that option `runtime_tmpdir` can be set and has effect. Applicable to onefile builds only.
    """

    pyi_builder.test_source(
        """
        import os
        import sys

        cwd = os.path.abspath(os.getcwd())
        runtime_tmpdir = os.path.abspath(sys._MEIPASS)

        # With --runtime-tmpdir=., we expect the application to unpack itself into cwd/_MEIXXXX directory.
        if os.path.dirname(runtime_tmpdir) != cwd:
            raise SystemExit(
                f'Expected sys._MEIPASS ({sys._MEIPASS}) to be under current working dir ({cwd}).'
            )
        """,
        # Set runtime-tmpdir to current working dir
        pyi_args=['--runtime-tmpdir', '.']
    )


@xfail(reason='Issue #3037 - all scripts share the same global vars')
def test_several_scripts1(pyi_builder_spec):
    """
    Verify each script has it's own global vars (original case, see issue #2949).
    """
    pyi_builder_spec.test_spec('several-scripts1.spec')


@xfail(reason='Issue #3037 - all scripts share the same global vars')
def test_several_scripts2(pyi_builder_spec):
    """
    Verify each script has it's own global vars (basic test).
    """
    pyi_builder_spec.test_spec('several-scripts2.spec')


def test_hyphenated_hiddenimport(pyi_builder):
    """
    Verify that a spec whose hiddenimports include a hyphenated module name is valid
    See issue #8591
    """
    pyi_builder.test_source(
        """
        print("hello!")
        """, pyi_args=['--hiddenimport', 'fake-hyphenated-module']
    )


@pytest.mark.win32
def test_pe_checksum(pyi_builder):
    import ctypes
    from ctypes import wintypes

    pyi_builder.test_source("print('hello')")
    exes = pyi_builder._find_executables('test_source')
    assert exes
    for exe in exes:
        # Validate the PE checksum using the official Windows API for doing so.
        # https://docs.microsoft.com/en-us/windows/win32/api/imagehlp/nf-imagehlp-mapfileandchecksumw
        header_sum = wintypes.DWORD()
        checksum = wintypes.DWORD()
        assert ctypes.windll.imagehlp.MapFileAndCheckSumW(
            ctypes.c_wchar_p(exe), ctypes.byref(header_sum), ctypes.byref(checksum)
        ) == 0

        assert header_sum.value == checksum.value


@onefile_only
def test_onefile_longpath(pyi_builder, tmp_path):
    """
    Verify that files with paths longer than 260 characters are correctly extracted from the onefile build.
    See issue #5615."
    """
    # Create data file with secret
    _SECRET = 'LongDataPath'
    src_path = tmp_path / 'data.txt'
    src_path.write_text(_SECRET, encoding='utf-8')
    # Generate long target filename/path; eight equivalents of SHA256 strings plus data.txt should push just the
    # _MEIPASS-relative path beyond 260 characters...
    dst_filename = os.path.join(*[32 * chr(c) for c in range(ord('A'), ord('A') + 8)], 'data.txt')
    assert len(dst_filename) >= 260
    # Name for --add-data
    add_data_arg = f"{src_path}:{os.path.dirname(dst_filename)}"
    pyi_builder.test_source(
        f"""
        import sys
        import os

        data_file = os.path.join(sys._MEIPASS, {str(dst_filename)!r})
        print("Reading secret from %r" % (data_file))
        with open(data_file, 'r') as fp:
            secret = fp.read()
        assert secret == {_SECRET!r}
        """,
        pyi_args=['--add-data', add_data_arg]
    )


@pytest.mark.win32
@pytest.mark.parametrize("icon", ["icon_default", "icon_none", "icon_given"])
def test_application_executable_has_manifest(pyi_builder, icon):
    """
    Verify that builds on Windows end up having manifest embedded. See issue #5624.
    This test was initially limited only to onefile builds, but as we now always embed manifest into executable, it
    now covers both builds.
    """
    from PyInstaller.utils.win32 import winmanifest
    from PyInstaller import PACKAGEPATH

    # Icon type
    if icon == 'icon_default':
        # Default; no --icon argument
        extra_args = []
    elif icon == 'icon_none':
        # Disable icon completely; --icon NONE
        extra_args = ['--icon', 'NONE']
    elif icon == 'icon_given':
        # Locate pyinstaller's default icon, and explicitly give it
        # via --icon argument
        icon_path = os.path.join(PACKAGEPATH, 'bootloader', 'images', 'icon-console.ico')
        extra_args = ['--icon', icon_path]
    # Build the executable...
    pyi_builder.test_source("""print('Hello world!')""", pyi_args=extra_args)
    # ... and ensure that it contains manifest
    exes = pyi_builder._find_executables('test_source')
    assert exes
    for exe in exes:
        manifest = winmanifest.read_manifest_from_executable(exe)
        assert manifest, "No manifest resources found!"


@pytest.mark.parametrize("append_pkg", [True, False], ids=["embedded", "sideload"])
def test_sys_executable(pyi_builder, append_pkg, monkeypatch):
    """
    Verify that sys.executable points to the executable, regardless of build mode (onedir vs. onefile) and the
    append_pkg setting (embedded vs. side-loaded CArchive PKG).
    """
    # Set append_pkg; taken from test_pyz_as_external_file
    import PyInstaller.building.build_main
    EXE = PyInstaller.building.build_main.EXE

    def MyEXE(*args, **kwargs):
        kwargs['append_pkg'] = append_pkg
        return EXE(*args, **kwargs)

    monkeypatch.setattr('PyInstaller.building.build_main.EXE', MyEXE)

    # Expected executable basename
    exe_basename = 'test_source'
    if is_win or is_cygwin:
        exe_basename += '.exe'

    pyi_builder.test_source(
        f"""
        import sys
        import os
        exe_basename = os.path.basename(sys.executable)
        assert exe_basename == {exe_basename!r}, "Unexpected basename(sys.executable): " + exe_basename
        """
    )


@pytest.mark.win32
def test_subprocess_in_windowed_mode(pyi_windowed_builder):
    """Test invoking subprocesses from a PyInstaller app built in windowed mode."""

    pyi_windowed_builder.test_source(
        fr"""
        from subprocess import PIPE, run
        from unittest import TestCase

        # Lazily use unittest's rich assertEqual() for assertions with builtin diagnostics.
        assert_equal = TestCase().assertEqual

        # Path to python interpreter
        python = {sys.executable!r}

        # Run with empty command to ensure interpreter works.
        run([python, "-c", ""], check=True)

        # Verify that stdin, stdout and stderr still work and haven't been muddled.
        p = run([python, "-c", "print('foo')"], stdout=PIPE, universal_newlines=True)
        assert_equal(p.stdout, "foo\n", p.stdout)

        p = run([python, "-c", r"import sys; sys.stderr.write('bar\n')"], stderr=PIPE, universal_newlines=True)
        assert_equal(p.stderr, "bar\n", p.stderr)

        p = run([python], input="print('foo')\nprint('bar')\n", stdout=PIPE, universal_newlines=True)
        assert_equal(p.stdout, "foo\nbar\n", p.stdout)
        """,
        pyi_args=["--windowed"]
    )


def test_package_entry_point_name_collision(pyi_builder):
    """
    Check when an imported package has the same name as the entry point script. Despite the obvious ambiguity, Python
    still handles this case fine and PyInstaller should too.
    """
    script = pathlib.Path(__file__).parent / 'data' / 'name_clash_with_entry_point' / 'matching_name.py'

    # Each file prints its filename and the value of __name__.
    expected = [
        ('matching_name.py', '__main__'),
        ('matching_name/__init__.py', 'matching_name'),
        ('matching_name/submodule.py', 'matching_name.submodule'),
    ]

    # Include a verification that unfrozen Python does still work.
    p = subprocess.run([sys.executable, str(script)], stdout=subprocess.PIPE, encoding="utf-8")
    assert re.findall("Running (.*) as (.*)", p.stdout) == expected

    pyi_builder.test_script(str(script))
    exe, = pyi_builder._find_executables("matching_name")
    p = subprocess.run([exe], stdout=subprocess.PIPE, encoding="utf-8")
    assert re.findall("Running (.*) as (.*)", p.stdout) == expected


@onedir_only
def test_contents_directory(pyi_builder):
    """
    Test the --contents-directory option, including changing it without --clean.
    """
    pyi_builder.test_source("", pyi_args=["--contents-directory", "foo"])
    exe, = pyi_builder._find_executables("test_source")
    bundle = pathlib.Path(exe).parent
    assert (bundle / "foo").is_dir()

    pyi_builder.test_source("", pyi_args=["--contents-directory", "é³þ³źć🚀", "--noconfirm"])
    assert not (bundle / "foo").exists()
    assert (bundle / "é³þ³źć🚀").is_dir()

    with pytest.raises(SystemExit, match='Invalid value "\\.\\." passed'):
        pyi_builder.test_source("", pyi_args=["--contents-directory", "..", "--noconfirm"])


@onedir_only
def test_legacy_onedir_layout(pyi_builder):
    """
    Test the --contents-directory=., which re-enables the legacy onedir layout.
    """
    pyi_builder.test_source(
        """
        import sys
        import os

        # NOTE: the paths set by bootloader (`sys._MEIPASS`, `__file__`) may end up using different separator than
        # paths set by the python interpreter itself (e.g., `sys.executable`) - for example, under msys2/mingw
        # python on Windows). Therefore, we normalize the separator via `os.path.normpath` before comparison.
        assert os.path.normpath(sys._MEIPASS) == os.path.dirname(sys.executable)
        assert os.path.normpath(os.path.dirname(__file__)) == os.path.dirname(sys.executable)
        """,
        pyi_args=["--contents-directory", "."]
    )


def test_spec_options(pyi_builder_spec, spec_dir, capsys):
    pyi_builder_spec.test_spec(
        spec_dir / "pyi_spec_options.spec",
        pyi_args=["--", "--optional-dependency", "email", "--optional-dependency", "gzip"]
    )
    exe, = pyi_builder_spec._find_executables("pyi_spec_options")
    p = subprocess.run([exe], stdout=subprocess.PIPE, encoding="utf-8")
    assert p.stdout == "Available dependencies: email gzip\n"

    capsys.readouterr()
    with pytest.raises(SystemExit) as ex:
        pyi_builder_spec.test_spec(spec_dir / "pyi_spec_options.spec", pyi_args=["--", "--help"])
    assert ex.value.code == 0
    assert "help blah blah blah" in capsys.readouterr().out

    with pytest.raises(SystemExit) as ex:
        pyi_builder_spec.test_spec(spec_dir / "pyi_spec_options.spec", pyi_args=["--", "--onefile"])
    assert "pyi_spec_options.spec: error: unrecognized arguments: --onefile" in capsys.readouterr().err
